/* ================================================================== */
/**
 *  @file iact.c
 *  @short CORSIKA interface for Imaging Atmospheric Cherenkov Telescopes etc.
 *
 *  @author  Konrad Bernloehr 
 *  $Date: 2006/04/07 11:38:39 $
 *  $Revision: 1.43 $
 *
 *  Copyright (C) 1997, 1998, 1999, 2001, 2002, 2005, 2006 Konrad Bernloehr. All rights reserved.
 *  Distribution and use of this software with the CORSIKA program is
 *  allowed and free. No redistribution separate of CORSIKA or of
 *  modified versions granted without permission. Modifications may,
 *  however, be distributed as patches to the original version.
 *  This software comes with no warranty.
 *
 *  Version 1.2.9
 *
 *  --------------------------------------------------------------------
 *
 *  This file implements a CORSIKA interface for the simulation
 *  of (3-D) arrays of Cherenkov telescopes.
 *  A whole array may be simulated in multiple instances with
 *  random offsets of each instance.
 *  For full use of this software additional files are required
 *  which are available now on request from Konrad Bernloehr
 *  (e-mail: Konrad.Bernloehr@mpi-hd.mpg.de).
 *  These additional files should be included in the same add-on
 *  package to CORSIKA which includes this file.
 *  A fallback mechanism is included to use the normal CORSIKA
 *  output of Cherenkov photon bunches instead of the dedicated
 *  output functions from the unavailable files. However, this
 *  fallback mechanism has important drawbacks: information about
 *  positions of telescopes are completely lost and no photon
 *  bunches are collected in memory because the collected bunches
 *  would never be written out. For those reasons you are adviced
 *  to obtain and use the additional software. 
 *
 *
 *  General comments on this file:
 *
 *  Routines provided in this file interface to recent versions
 *  of the CORSIKA air shower simulation program. Modifications to
 *  CORSIKA have been kept as simple as possible and the existing
 *  routines for production of Cherenkov light have been largely
 *  maintained. Setup of the telescope systems to be simulated is
 *  via the usual CORSIKA input file (the syntax of which has been
 *  extended by a few additional keywords). These telescope systems
 *  can be randomly scattered several times within a given area.
 *  All treatment whether a bunch of photons hits a telescope is done
 *  by the routines in this file. Photon bunches are kept in main
 *  memory until the end of the event. This might be a limitation
 *  when simulating large showers / many telescopes / many systems of
 *  telescopes on a computer with little memory. An option to store
 *  photon bunches in a temporary file has, therefore, been included.
 *  After the end of an event in CORSIKA all photon bunches (sorted
 *  by system and telescope) are written to a data file in the
 *  'eventio' portable data format also used for CRT and HEGRA CT data.
 *  All CORSIKA run/event header/trailer blocks are also written to
 *  this file.
 *
*/

/* ==================================================================== */

#define IACT_ATMEXT_VERSION "1.35 (2006-04-06)"

#ifndef MAX_IO_BUFFER
# define MAX_IO_BUFFER 200000000
#endif

#if defined(IACTEXT)
# define EXTENDED_TELOUT 1
#elif defined(EXTENDED_TELOUT)
# define IACTEXT 1
#endif

/* Compile with '-DNO_EVENTIO' if eventio functions are not available. */
#ifndef NO_EVENTIO
# define HAVE_EVENTIO_FUNCTIONS 1
#endif

/* Without 'eventio' functions saving the photon bunches is left to CORSIKA. */
/* If you want both kinds of output, compile with '-DCORSIKA_SAVES_PHOTONS' */
#ifndef CORSIKA_SAVES_PHOTONS
# ifndef HAVE_EVENTIO_FUNCTIONS
#  define CORSIKA_SAVES_PHOTONS 1
# endif
#endif

/* Compile with '-DNO_PIPE' if piping of output directly to other programs */
/* is not wanted at all. To actually use pipes, start 'TELFIL' with a '|'. */
#ifndef NO_PIPE
# define PIPE_OUTPUT 1
#endif

/* The Cherenkov light longitudinal distribution from CORSIKA might be  */
/* in either of two ways: an emission profile or an integrated profile. */
/* When neither INTEGRATED_LONG_DIST nor EMISSION_LONG_DIST is declared */
/* we try to find it out dynamically. */
#ifndef INTEGRATED_LONG_DIST
# ifndef EMISSION_LONG_DIST
#  define UNKNOWN_LONG_DIST 1
# endif
#endif




#ifdef HAVE_EVENTIO_FUNCTIONS
#include "initial.h"      /* This file includes others as required. */
#include "io_basic.h"     /* This file includes others as required. */
#include "mc_tel.h"
#ifndef NO_FILEOPEN
#include "fileopen.h"
#endif
#else
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#endif

#include "sampling.h"
#include "straux.h"


/* ------------ CORSIKA version-specific definitions -------------- */

#ifndef CORSIKA_VERSION
# define CORSIKA_VERSION 6500 /* the version that should be matched */
#endif

#if (CORSIKA_VERSION < 5900)
# define PRMPAR_SIZE 14
#else
# define PRMPAR_SIZE 17
#endif

/** Type for CORSIKA floating point numbers remaining REAL*4 */
typedef float cors_real_t;
/** Type for many CORSIKA numbers has changed to REAL*8 with version 5.901. */
#if (CORSIKA_VERSION < 5901)
typedef float cors_real_now_t;
#else
typedef double cors_real_now_t;
#endif
/** Type for CORSIKA numbers which were REAL*4 but changed to REAL*8 at 5.900 */
#if (CORSIKA_VERSION < 5900)
typedef float cors_real_dbl_t;
#else
typedef double cors_real_dbl_t;
#endif
/** Type for CORSIKA numbers which were already REAL*8 */
typedef double cors_dbl_t;

/* ----------- Standard C or Unix-specific include files ---------- */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#ifdef PIPE_OUTPUT
#include <ctype.h>
#endif

/* The following definitions are left over from a rare memory problem */
/* that turned out to be due to a RAM hardware problem. */
/* Should be removed after more testing. */

#ifndef EXTRA_MEM
# define EXTRA_MEM 0
#endif
#define EXTRA_MEM_1 EXTRA_MEM
#define EXTRA_MEM_2 EXTRA_MEM
#define EXTRA_MEM_3 EXTRA_MEM
#define EXTRA_MEM_4 EXTRA_MEM
#define EXTRA_MEM_5 EXTRA_MEM
#define EXTRA_MEM_6 EXTRA_MEM
#define EXTRA_MEM_7 EXTRA_MEM
#define EXTRA_MEM_8 EXTRA_MEM
#define EXTRA_MEM_9 EXTRA_MEM
#define EXTRA_MEM_10 EXTRA_MEM
#define EXTRA_MEM_11 EXTRA_MEM
#define EXTRA_MEM_12 EXTRA_MEM

/* -------------- Use the RMMAR[D] random number generator ----------- */

#if ( CORSIKA_VERSION >= 6145 )
extern void rmmard_(double *, int *, int *);
#else
extern void rmmar_(float *, int *, int *);
#endif

static double rndm(int dummy);

/** @short Random number interface using sequence 4 of CORSIKA. */
static double rndm(int dummy)
{
   int num = 1;
   int seq = 4;     /* Use sequence number 4 of RMMAR[D] */
#if ( CORSIKA_VERSION >= 6145 )
   static double rtmp[10];
   rmmard_(rtmp,&num,&seq); /* Call Fortran subroutine */
#else
   static float rtmp[10];
   rmmar_(rtmp,&num,&seq); /* Call Fortran subroutine */
#endif
   return rtmp[0];
}

double iact_rndm(int dummy);

double iact_rndm(int dummy)
{
   return rndm(dummy);
}

/* -------------------------------------------------------------- */

#ifndef Nint
#define Nint(x) ((x)>0?(int)((x)+0.5):(int)((x)-0.5))
#endif
#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

/** The CORSIKA version actually running. */
int corsika_version = (CORSIKA_VERSION);

/** Maximum number of telescopes (or other detectors) per array */
#define MAX_ARRAY_SIZE 1000
/** Position and size definition of fiducial spheres. */
static double xtel[MAX_ARRAY_SIZE], ytel[MAX_ARRAY_SIZE],
              ztel[MAX_ARRAY_SIZE], rtel[MAX_ARRAY_SIZE];
static double raise_tel;  /**< Non-zero if any telescope has negative z */
static double rmax = 0.;  /**< Max. radius of telescopes */
static double dmax = 0.;  /**< Max. distance of telescopes in (x,y) */
static int ntel = 0;      /**< Number of telescopes set up */
static int nsys = 1;      /**< Number of arrays */

/* Speed of light used to correct for the delay between observation level */
/* and the actual telescope elevation. A constant should be good enough. */ 
static double airlightspeed = 29.9792458/1.0002256; /**< [cm/ns] at H=2200 m */

/** The maximum core offset of array centres in circular distribution. */
static double core_range;
/** The maximum core offsets in x,y for rectangular distribution. */
static double  core_range1, core_range2;
static double impact_offset[2]; /**< Offset of impact position of charged primaries */
static int impact_correction = 1; /**< Correct impact position if non-zero. */

/** The central value of the allowed ranges in theta and phi. */
static double theta_central, phi_central, off_axis;

/* The number of events for which full output is 'printed' can be limited. */
/* You can control that with the TELFIL option for the output file. */
/* Defaults correspond to 'TELFIL iact.dat:10:100:1' */
static int count_print_tel = 0, count_print_evt = 0;
static int max_print_tel = 10, max_print_evt = 100;
static int skip_print = 1, skip_print2 = 100;
static int skip_off2 = 1;

/** The name of the file providing parameters for importance sampling. */
static char *sampling_fname;
/** The name of the output file for eventio format data. */
static char *output_fname;
/** The largest block size in the output data, which must hold all
    photons bunches of one array. */
static size_t max_io_buffer = MAX_IO_BUFFER;
#ifdef HAVE_EVENTIO_FUNCTIONS
/* Data structures are defined in mc_tel.h */
static IO_BUFFER *iobuf;
static void ioerrorcheck ();
#else
/* Data structures: */

/** A photon (or photoelectron) bunch in full mode. */
struct bunch
{
   float photons; /**< Number of photons in bunch */
   float x, y;    /**< Arrival position relative to telescope (cm) */
   float cx, cy;  /**< Direction cosines of photon direction */
   float ctime;   /**< Arrival time (ns) */
   float zem;     /**< Height of emission point above sea level (cm) */
   float lambda;  /**< Wavelength in nanometers or 0 (negative if scattered) */
};

/** A photon (or photoelectron) bunch in compact mode. */
struct compact_bunch
{
   short photons; /**< ph*100 */
   short x, y;    /**< x,y*10 (mm) */
   short cx, cy;  /**< cx,cy*30000 */
   short ctime;   /**< ctime*10 (0.1ns) after subtracting offset */
   short log_zem; /**< log10(zem)*1000 */
   short lambda;  /**< (nm) or 0 (negative for scattered light) */
};

/** A linked list of text strings. */
struct linked_string
{
   char *text;
   struct linked_string *next;
};

#endif

#define NBUNCH 1000           /**< Memory allocation step size for bunches */
#define INTERNAL_LIMIT 100000 /**< Start external storage after so many bunches */
#define EXTERNAL_STORAGE 1    /**< Enable external temporary bunch storage */

/** A structure describing a detector and linking its photons bunches to it. */
struct detstruct
{
   double x, y, x0, y0, z0, r0;
   double sampling_area;
   double r, dx, dy;
   int geo_type;
   int sens_type;
   int class;
   int iarray;
   int idet;
   int bits;
   double photons;
   struct bunch *bunch;
   struct compact_bunch *cbunch;
   int available_bunch;
   int next_bunch;
   char ext_fname[60];
   int external_bunches;
};

struct gridstruct
{
   int ndet, idet;
   struct detstruct **detectors;
};

#define MAX_CLASS 1

#ifdef IACTEXT
static struct bunch *particles;
int particles_space;
int particles_stored;
#endif

/** The largest number of photon bunches kept in main memory before
    attempting to flush them to temporary files on disk. */
static int max_internal_bunches = INTERNAL_LIMIT;
static int narray;
static int *ndet;
static int nevents, event_number;
static int do_print;
static struct detstruct **detector;
static int det_in_class[MAX_CLASS];
static double *xoffset, *yoffset, *weight;
static struct gridstruct *grid;
static double grid_x_low, grid_y_low, grid_x_high, grid_y_high;
static int grid_nx, grid_ny, grid_elements;

static struct linked_string corsika_inputs = { "* CORSIKA inputs:", NULL };

static int televt_done;
static double obs_height, toffset;
static double energy, theta_prim, phi_prim;
#ifdef SHOW_ANGLE
static double cx_prim, cy_prim, cz_prim;
#endif
static int primary;
static double first_int;

#if defined(IACTEXT) || defined(EXTENDED_TELOUT)
static int use_compact_format = 0;
#else
static int use_compact_format = 1;
#endif

static double all_photons, all_photons_run;
static double all_bunches, all_bunches_run;
static long stored_bunches;
static double lambda1, lambda2;

#define GRID_SIZE 1000  /**< unit: cm */


/* =============================================================== */

/* FORTRAN called functions (beware changes of parameter types !!) */
/* The additional character string lengths for name in telfil_ and */
/* line in tellni_ are not used because compiler-dependent.        */
void telfil_(char *name);
void telsmp_(char *name);
void telshw_(void);
void telinf_(int *itel, 
   double *x, double *y, double *z, double *r, int *exists);
void tellni_(char *line, int *llength); 
void telrnh_(cors_real_t runh[273]);
void telrne_(cors_real_t rune[273]);
void telasu_(int *n, cors_real_dbl_t *dx, cors_real_dbl_t *dy);
void telset_(cors_real_now_t *x, cors_real_now_t *y, 
   cors_real_now_t *z, cors_real_now_t *r);
void televt_(cors_real_t evth[273], cors_real_dbl_t prmpar[PRMPAR_SIZE]);
int telout_(cors_real_now_t *bsize, cors_real_now_t *wt, 
   cors_real_now_t *px, cors_real_now_t *py, cors_real_now_t *pu, 
   cors_real_now_t *pv, cors_real_now_t *ctime, 
   cors_real_now_t *zem, cors_real_now_t *lambda
#ifdef EXTENDED_TELOUT
   , double *temis, double *penergy, double *amass, double *charge
#endif
   );
#ifdef IACTEXT
void telprt_ (cors_real_t* datab, int *maxbuf);
#endif
void tellng_ (int *type, double *data, int *ndim, int *np, 
   int *nthick, double *thickstep);
void telend_(cors_real_t evte[273]);

/* CORSIKA function called from this module: */
extern double heigh_ (double *thickness);

/* C called functions (parameter types are always checked) */
static void iact_param (char *text);
static int photon_hit (struct detstruct *det, double x, double y,
   double cx, double cy, double sx, double sy, double photons,
   double ctime, double zem, double lambda);
static int in_detector (struct detstruct *det, double x, double y, 
   double sx, double sy);
static int set_random_systems (double theta, double phi, 
   double thetaref, double phiref, double offax,
   double E, int primary, int volflag);

void sample_offset (char *sampling_fname, double core_range, 
   double theta, double phi, 
   double thetaref, double phiref, double offax, 
   double E, int primary,
   double *xoff, double *yoff, double *sampling_area);

static double ush, vsh, wsh;
static double ushc, vshc, wshc;

extern double refidx_ (double *height);
extern double rhof_ (double *height);

/* ================================================================== */

#ifdef CATCH_SIGNALS
#include <signal.h>
#if defined(OS_ULTRIX)
void stop_signal_function (int isig, int code, struct sigcontext *scp);
#else
void stop_signal_function (int isig);
#endif

/** Catching signals and then exit makes sense with dmalloc, for example. */
#if defined(OS_ULTRIX)
void stop_signal_function (int isig, int code, struct sigcontext *scp)
#else
void stop_signal_function (int isig)
#endif
{
   if ( isig >= 0 )
   {
#if defined(OS_ULTRIX)
      fprintf(stderr,"Received signal %d (code %d)\n",isig,code);
#else
      fprintf(stderr,"Received signal %d\n",isig);
#endif
   }
   
   signal(SIGINT,SIG_DFL);
   signal(SIGTERM,SIG_DFL);
   
   /* Free all allocated memory. With memory debugging tools this helps to */
   /* reassure that no memory leaks or other memory allocation problems exist. */
   
   /* Free grid from previous showers */
   
   if ( grid != NULL )
   {
      int ix, iy;
      for (iy=0; iy<grid_ny; iy++)
         for (ix=0; ix<grid_nx; ix++)
         {
            struct gridstruct *gc;
            if ( iy*grid_nx+ix < 0 || iy*grid_nx+ix >= grid_elements )
            {
               fprintf(stderr,
              "Grid bounds exceeded: ix=%d, iy=%d, i=%d (nx=%d, ny=%d, n=%d)\n",
                  ix, iy, iy*grid_nx+ix, grid_nx, grid_ny, grid_elements);
            }
            gc = &grid[iy*grid_nx+ix];
            if ( gc->detectors != NULL )
            {
               free(gc->detectors);
               gc->detectors = NULL;
            }
            gc->ndet = gc->idet = 0;
         }
      free(grid);
      grid = NULL;
   }
    
   if ( detector != 0 )
   {
      int iarray;
      for (iarray=0; iarray<narray; iarray++)
      {
         if ( detector[iarray] != NULL )
         {
            int idet;
	    for (idet=0; idet<ndet[iarray]; idet++)
            {
               if ( detector[iarray][idet].cbunch != NULL )
               {
                  free(detector[iarray][idet].cbunch);
                  detector[iarray][idet].cbunch = NULL;
               }
               if ( detector[iarray][idet].bunch != NULL )
               {
                  free(detector[iarray][idet].bunch);
                  detector[iarray][idet].bunch = NULL;
               }
            }
            free(detector[iarray]);
            detector[iarray] = NULL;
         }
      }
      free(detector);
      detector = NULL;
   }
   
   if ( ndet != NULL )
   {
      free(ndet);
      ndet = NULL;
   }

   /* Search end of linked list */
   {
      struct linked_string *xl;
      for (xl=&corsika_inputs; xl != NULL; )
      {
         struct linked_strings *xln = xl->next;
         xl->next = NULL;
         if ( xl != &corsika_inputs )
         {
            if ( xl->text != NULL )
            {
               free(xl->text);
               xl->text = NULL;
            }
            free(xl);
         }
         xl = xln;
      }
   }
   
   if ( output_fname != NULL )
   {
      free(output_fname);
      output_fname = NULL;
   }
   
   if ( xoffset != NULL )
   {
      free(xoffset);
      xoffset = NULL;
   }

   if ( yoffset != NULL )
   {
      free(yoffset);
      yoffset = NULL;
   }
   
   if ( weight != NULL )
   {
      free(weight);
      weight = NULL;
   }
   
   initpath(NULL); /* Two allocations remain there */
   
   if ( iobuf != NULL )
   {
      if ( iobuf->output_file != NULL )
         fileclose(iobuf->output_file);
      iobuf->output_file = NULL;
      free_io_buffer(iobuf);
      iobuf = NULL;
   }

   exit(1);
}

#endif

#ifdef HAVE_EVENTIO_FUNCTIONS
void ioerrorcheck (void)
{
   static size_t ioerrcnt = 0;
   
   if ( ++ioerrcnt > 100 )
   {
      fprintf(stderr,"\n\nFatal error:\n"
          "Too many errors on the IACT output file were detected.\n"
          "It makes little sense to continue with CORSIKA when no output\n"
          "can be written or the output file is badly corrupted.\n"
          "This situation could arise when your disk is full or the output is\n"
          "piped into a detector simulation program and that happened to fail.\n");
      exit(9);
   }
}
#endif

/* ------------------------------ telfil_ --------------------------- */
/**
 * @short Define the output file for photon bunches hitting the telescopes.
 *
 * This function is called when the 'TELFIL' keyword is present in
 * the CORSIKA input file.
 *
 * @verbatim
 * The 'file name' parsed is actually decoded further:
 *    Apart from the leading '+' or '|' or '+|' the TELFIL argument
 *    may contain further bells ans whistles:
 *    If the supplied file name contains colons, they are assumed to 
 *    separate appended numbers with the following meaning:
 *      #1:  number of events for which the photons per telescope are shown
 *      #2:  number of events for which energy, direction etc. are shown
 *      #3:  every so often an event is shown (e.g. 10 -> every tenth event).
 *      #4:  every so often the event number is shown even if #1 and #2 ran out.
 *      #5:  offset for #4 (#4=100, #5=1: show events 1, 101, 201, ...)
 *      #6:  the maximum number of photon bunches before using external storage
 *      #7:  the maximum size of the output buffer in Megabytes.
 *    Example: name = "iact.dat:5:15:10"
 *       name becomes "iact.dat"
 *       5 events are fully shown
 *       15 events have energy etc. shown
 *       Every tenth event is shown, i.e. 10,20,30,40,50 are fully shown
 *       and events number 60,...,150 have their energies etc. shown.
 *       After that every shower with event number divideable by 1000 is shown.
 *    Note: No spaces inbetween! CORSIKA input processing truncates at blanks.
 * @endverbatim
 *
 * @param  name    Output file name.
 * 		   Note: A leading '+' means: use non-compact format
 * 		   A leading '|' (perhaps after '+') means that the
 * 		   name will not be interpreted as the name of
 * 		   a data file but of a program to which the
 * 		   'eventio' data stream will be piped (i.e. that
 * 		   program should read the data from its standard
 * 		   input.
 * @return (none)
 *
*/

void telfil_ (char *name)
{
   int i, l;
   char *s;

   if ( (s=strchr(name,':')) != NULL )
   {
      *s = '\0';
      if ( max_print_tel == 0 )
      {
         max_print_tel = max_print_evt = atoi(s+1);
         skip_print = 1; /* Well, actually 1 means don't skip any event */
      }
      if ( (s=strchr(s+1,':')) != NULL )
      {
	 if ( (max_print_evt = atoi(s+1)) < max_print_tel )
	    max_print_evt = max_print_tel;
	 if ( (s=strchr(s+1,':')) != NULL )
         {
	    if ( (skip_print = atoi(s+1)) <= 0 )
	       skip_print = 1;
            if ( (s=strchr(s+1,':')) != NULL )
            {
	       if ( (skip_print2 = atoi(s+1)) <= 0 )
	          skip_print2 = 1000;
               if ( (s=strchr(s+1,':')) != NULL )
               {
	          if ( (skip_off2 = atoi(s+1)) <= 0 )
	             skip_off2 = 0;
                  if ( (s=strchr(s+1,':')) != NULL )
                  {
	             if ( (max_internal_bunches = atoi(s+1)) <= 1000 )
	                max_internal_bunches = INTERNAL_LIMIT;
                     if ( (s=strchr(s+1,':')) != NULL )
                     {
                        /* Maximum size of I/O buffer in 10^6 bytes units. */
                        if ( atoi(s+1) >= 1 && atoi(s+1) < 64000 )
                           max_io_buffer = ((size_t) atoi(s+1)) * 1000000;
                     }
                  }
               }
            }
	 }
      }
   }

   /* The rest of the string should be the real file name */
   l = strlen(name);
   if ( l > 99 )
   {
      fprintf(stderr,
         "\n Output file name of length %d truncated to 99 characters.\n\n",l);
      l = 99;
   }
#ifdef REMOVE_PADDING_BLANKS
   for (i=l; i>0; i--)
      if ( name[i-1] != ' ' && name[i-1] != '\n' && name[i-1] != '\0' )
         break;
#else
   i = l;
#endif
   name[i] = '\0';
   if ( *name == '+' && i > 0 )
   {
      name++; i--;
      use_compact_format = 0;
   }
#if !defined(IACTEXT) && !defined(EXTENDED_TELOUT)
   else
      use_compact_format = 1;
#endif
   if ( i <= 0 )
   {
      fprintf(stderr,"Missing or invalid filename for Cherenkov photons.\n");
      exit(1);
   }
   output_fname = malloc((size_t)i+1+EXTRA_MEM_1);
   strncpy(output_fname,name,(size_t)i);
   output_fname[i] = '\0';
   if ( strncmp(output_fname,"/dev/null",(size_t)9) == 0 )
      strcpy(output_fname,"/dev/null");
#ifndef HAVE_EVENTIO_FUNCTIONS
   fprintf(stderr,"\n Output filename for Cherenkov photons ignored.\n\n");
#endif
}


/* ----------------------------- telsmp_ --------------------------- */
/**
 *  @short Set the file name with parameters for importance sampling.
 *
 *  Note that the TELSAMPLE parameter is not processed by CORSIKA itself
 *  and thus has to be specified through configuration lines like
@verbatim
IACT TELSAMPLE filename
*(IACT) TELSAMPLE filename
@endverbatim
where the first form requires a CORSIKA patch and the second
would work without that patch (but then only with uppercase file names).
 */

void telsmp_ (char *name)
{
   size_t i, len = 0;
   if ( name == 0 )
      return;
   len = strlen(name);
   for (i=0; i<len; i++)
      if ( !isalnum(name[i]) && name[i] != '_' && name[i] != '.' && 
           name[i] != '-' && name[i] != '/' )
      {
         fprintf(stderr,"\n Invalid file name for core offset sampling parameters.\n\n");
         return; 
      }
   sampling_fname = malloc((size_t)len+1+EXTRA_MEM_1);
   strcpy(sampling_fname,name);
}

/* ------------------------------ telshw_ --------------------------- */
/**
 *  @short Show what telescopes have actually been set up.
 *
 *  This function is called by CORSIKA after the input file is read.
 *
*/

void telshw_ ()
{
   int i;
   fflush(NULL);
#ifdef HAVE_EVENTIO_FUNCTIONS
   fprintf(stdout,"\n Telescope output file: '%s'\n",output_fname);
#endif
   fprintf(stdout,"\n Number of simulated telescopes: %d\n",ntel);
   if ( raise_tel != 0. )
      fprintf(stdout," All telescopes are raised by %4.2f m\n",
         0.01*raise_tel);
   for (i=0; i<ntel; i++)
      fprintf(stdout,
      "    Telescope %d at x=%6.2f m, y=%6.2f m, z=%6.2f m with r=%4.2f m\n",
         i+1,xtel[i]/100.,ytel[i]/100.,ztel[i]/100.,rtel[i]/100.);          
   fprintf(stdout,"\n");
   fprintf(stdout," Number of telescope arrays simulated: %d\n",nsys);
   if ( core_range2 <= 0. )
      fprintf(stdout,
      " Array centers are at random offsets within %5.2f m radius from core\n",
         core_range/100.);
   else
      fprintf(stdout,
      " Array centers are at random offsets within %5.2f m by %5.2f m from core\n",
         core_range1/100.,core_range2/100.);
   if ( sampling_fname != NULL )
      fprintf(stdout,
      " The distribution of core offsets within this area is controlled through the\n"
      " parameter file '%s'.\n"
      " All generated events are recorded with area weights.\n",
      sampling_fname);
   fprintf(stdout,"\n");
   fflush(stdout);
}

/* --------------------------- telinf_ ------------------------------ */
/**
 * @short Return information about configured telescopes back to CORSIKA
 *
 * @param  itel     number of telescope in question
 * @param  x, y, z  telescope position [cm]
 * @param  r	    radius of fiducial volume [cm]
 * @param  exists   telescope exists
 *
*/

void telinf_ (int *itel, 
   double *x, double *y, double *z, double *r, int *exists)
{
   if ( *itel <= 0 || *itel > ntel )
   {
      *exists = 0;
      *x = *y = *z = *r = 0.;
   }
   else
   {
      *exists = 1;
      *x = xtel[*itel-1];
      *y = ytel[*itel-1];
      *z = ztel[*itel-1];
      *r = rtel[*itel-1];
   }
}


/* --------------------------- telrnh_ ------------------------------ */
/**
 *  @short Save aparameters from CORSIKA run header.
 *
 *  Get relevant parameters from CORSIKA run header block and
 *  write run header block to the data output file.
 *
 *  @param  runh CORSIKA run header block
 *  @return (none)
 * 
*/

void telrnh_ (cors_real_t runh[273])
{
#ifdef HAVE_EVENTIO_FUNCTIONS
   struct stat st;
#endif

#ifdef CATCH_SIGNALS
   /* Catch INTerrupt and TERMinate signals to stop program */
   signal(SIGINT,stop_signal_function);
   signal(SIGTERM,stop_signal_function);
#endif

   fflush(NULL);
   fprintf(stderr,"\n Using IACT/ATMEXT package version %s for CORSIKA %5.3f\n\n",
      IACT_ATMEXT_VERSION, (double)(CORSIKA_VERSION)/1000.);
   fflush(NULL);

   corsika_version = (int)(runh[3]*1000.+0.5);

   /* Check if versions match (up to 2 decimals) */
   if ( corsika_version != (CORSIKA_VERSION) )
   {
      /* Check for critical version numbers with parameter type changes */
      if ( (corsika_version >= 5900 && (CORSIKA_VERSION) <  5900) ||
           (corsika_version >= 6400 && (CORSIKA_VERSION) <  6400) ||
           (corsika_version <  6400 && (CORSIKA_VERSION) >= 6400) )
      {
         fprintf(stderr,
          "\nCORSIKA version %5.3f does not match IACT interface for version %5.3f\n",
            (double)runh[3], (double)(CORSIKA_VERSION)/1000.);
         exit(1);
      }
      /* Unless more things change, it should work but caution the user */
      fprintf(stdout,
       "\n CORSIKA version is %5.3f but IACT interface was adapted to version %5.3f.\n",
          (double)runh[3],(double)CORSIKA_VERSION/1000.);
      fprintf(stdout,
       " You might want to check that parameters passed are of matching types.\n\n");
      fflush(NULL);
   }

#ifdef HAVE_EVENTIO_FUNCTIONS
   /* Allocate I/O buffer for event data */
   if ( iobuf == 0 )
   {
      if ( (iobuf = allocate_io_buffer(100000)) == (IO_BUFFER *) NULL )
      {
         fflush(NULL);
         fprintf(stderr,"No I/O buffer can be allocated.\n");
         exit(1);
      }

      if ( getenv("CORSIKA_IO_BUFFER") != 0 )
      {
         char *s = getenv("CORSIKA_IO_BUFFER");
         long nb = atol(s);
         long bs = 1;
         if ( strstr(s,"M") != NULL || strstr(s,"MB") != NULL || nb < 64000 )
            bs = 1000000;
         if ( nb*bs > max_io_buffer )
            max_io_buffer = nb*bs;
      }

      /* If you ever get a message that the I/O buffer could not be     */
      /* extended but you think you have enough memory in your machine, */
      /* increase the following maximum (in units of bytes):            */
      if ( max_io_buffer != MAX_IO_BUFFER )
         printf(" Using a maximum output buffer size of %d MB for the IACT module.\n",
             max_io_buffer/1000000);

      iobuf->max_length = max_io_buffer;

      if ( output_fname == NULL )
         output_fname = "telescope.dat";

#ifdef PIPE_OUTPUT
      /* Output can be to a pipe if the first (remaining) character is a '|'. */
      if ( output_fname[0] == '|' )
      {
      	 int i, l;
	 l = strlen(output_fname);
	 for (i=1; i<l; i++)
	 {
	    /* Check that no shell metacharacters are in the string */
	    if ( !isalnum(output_fname[i]) &&
	         strchr(" \t=_-+.,:%%\'\"/@",output_fname[i]) == NULL )
	    {
	       fflush(NULL);
	       fprintf(stderr,"Invalid character '%c' in output pipe string.\n",
	          output_fname[i]);
	       exit(1);
	    }
	 }
      	 iobuf->output_file = popen(output_fname+1,"w");
      }
      else
#endif
      {
      	 /* Open output file in append mode to make sure that existing */
      	 /* data is not overwritten. */
#ifndef NO_FILEOPEN
         iobuf->output_file = fileopen(output_fname,"a");
#else
#ifdef __USE_LARGEFILE64
         /* With explicit support for large files on 32-bit machines. */
         iobuf->output_file = fopen64(output_fname,"a");
#else
         /* Support for large files is implicit or not available. */
         iobuf->output_file = fopen(output_fname,"a");
#endif
#endif
      }

      if ( iobuf->output_file == NULL )
      {
         fflush(NULL);
         perror(output_fname);

         if ( errno == EACCES || errno == ESPIPE )
         {
            fprintf(stderr,
             "In most cases this error message means that you have the data file named\n");
            fprintf(stderr,"'%s' left over from a previous run.\n",
               output_fname);
            fprintf(stderr,
             "You may append new data to this file if you enable write permissions\n");
            fprintf(stderr,"to the data file before restarting this program.\n");
            fprintf(stderr,"Otherwise rename or remove the data file first.\n");
         }
         exit(1);
      }
   }
#else
   fflush(NULL);
   fprintf(stderr,"\n You don't have IACT Cherenkov light output functions.\n");
   fprintf(stderr," The fallback solution is to use CORSIKA output.\n");
   fprintf(stderr," Note that CORSIKA output files contain no information on\n");
   fprintf(stderr," telescope positions or random offsets of telescope systems.\n\n");
#endif
   { int nht = runh[4];
     if ( nht>0 && nht <= 10 )
       obs_height = runh[4+nht];
     else
       obs_height = -100.;
   }
#ifdef HAVE_EVENTIO_FUNCTIONS
   if ( write_tel_block(iobuf,IO_TYPE_MC_RUNH,(int)runh[1],runh,273) != 0 )
      ioerrorcheck();
   if ( write_input_lines(iobuf,&corsika_inputs) != 0 )
      ioerrorcheck();
   fflush(iobuf->output_file);

   /* Setting file to read-only is delayed until here, so that */
   /* it also works with gzip/bzip2 compressed files. */
   if ( fstat(fileno(iobuf->output_file),&st) != 0 )
      perror("Error getting access mode of output file");
#ifdef S_ISFIFO
   else if ( !S_ISFIFO(st.st_mode) )
#else
   else if ( !((st.st_mode & S_IFMT) == S_IFIFO) )
#endif
   {
      /* Disable writing by other programs. User must enable write */
      /* bit by hand if appending of other data is wanted later on. */
      fchmod(fileno(iobuf->output_file),st.st_mode&~(unsigned)0333);
   }
   else if ( stat(output_fname,&st) != 0 )
      perror("Error getting access mode of output file");
   else
   {
      /* Disable writing by other programs. User must enable write */
      /* bit by hand if appending of other data is wanted later on. */
      chmod(output_fname,st.st_mode&~(unsigned)0333);
   }
#endif

   all_photons_run = 0.;
   all_bunches_run = 0.;

}

/* ---------------------------- tellni_ -------------------------- */
/**
 *  @short Keep a record of CORSIKA input lines.
 *
 *  Add a CORSIKA input line to a linked list of strings which
 *  will be written to the output file in eventio format right
 *  after the run header.
 *
 *  @param  line     input line (not terminated)
 *  @param  llength  maximum length of input lines (132 usually)
 *
*/

void tellni_ (char *line, int *llength)
{
   int len = *llength, i;
   struct linked_string *xl, *xln;
   char myline[512];

   /* Skip backwards over trailing blanks. */
   for (i=len-1; i>0; i--)
      if ( line[i] == ' ' )
      	 len--;
      else
      	 break;

   /* We may have some configuration concerning this module */
   if ( len < sizeof(myline) )
   {
      strncpy(myline,line,len);
         myline[len] = '\0';
      if ( strncmp(myline,"IACT ",5) == 0 )
         iact_param(myline+5);
      else if ( strncmp(myline,"* (IACT) ",9) == 0 )
         iact_param(myline+9);
      else if ( strncmp(myline,"*(IACT) ",8) == 0 )
         iact_param(myline+8);
   }

   /* Search end of linked list */
   for (xl=&corsika_inputs; xl->next != NULL; xl=xl->next)
      ;

   /* Append new line */
   if ( len > 0 )
      if ( (xln = calloc((size_t)1+EXTRA_MEM_2,sizeof(struct linked_string))) != NULL )
      {
         if ( (xln->text = malloc((size_t)(len+1+EXTRA_MEM_3))) != NULL )
         {
            if ( len > 0 )
      	       strncpy(xln->text,line,(size_t)len);
	    xln->text[len] = '\0';
	    xl->next = xln;
         }
      }
}

/* ----------------------- is_on ------------------ */

static int is_on (char *word)
{
   if ( strcasecmp(word,"on") == 0 ||
        strcasecmp(word,"yes") == 0 ||
        strcasecmp(word,"Y") == 0 ||
        strcasecmp(word,"True") == 0 ||
        strcasecmp(word,"T") == 0 ||
        strcasecmp(word,"1") == 0 )
      return 1;
   else
      return 0;
}

/* ----------------------- is_off ------------------ */

static int is_off (char *word)
{
   if ( strcasecmp(word,"off") == 0 ||
        strcasecmp(word,"no") == 0 ||
        strcasecmp(word,"N") == 0 ||
        strcasecmp(word,"False") == 0 ||
        strcasecmp(word,"F") == 0 ||
        strcasecmp(word,"0") == 0 )
      return 1;
   else
      return 0;
}

/* ------------------------- iact_param ------------------------- */
/** @short Processing of IACT module specific parameters in Corsika input.
 *
 *  @param text Text following the IACT keyword on the input line.
 *
 */

static void iact_param (char *text)
{
   char word[128], word2[128];
   int ipos = 0;
   char *s = strchr(text,'#');
   /* If there are comments appended, strip them off */
   if ( s != NULL )
   {
      int l;
      *s = '\0';
      if ( strlen(text) == 0 )
         return;
      for (l=strlen(text)-1; l>=0; l--)
         if ( text[l] == ' ' || text[l] == '\t' )
            text[l] = '\0';
         else
            break;
   }
   if ( getword(text,&ipos,word,sizeof(word)-1,' ','\n') <= 0 )
      return;
   if ( strcasecmp(word,"TELFIL") == 0 )
      telfil_(text+ipos);
   else if ( strcasecmp(word,"TELSAMPLE") == 0 )
      telsmp_(text+ipos);
   else if ( strcasecmp(word,"impact_correction") == 0 )
   {
      if ( getword(text,&ipos,word,sizeof(word)-1,' ','\n') <= 0 )
         return;
      if ( is_on(word) )
         impact_correction = 1;
      else if ( is_off(word) )
         impact_correction = 0;
   }
   else if ( strcasecmp(word,"print_events") == 0 )
   {
      /* See the first five colon separated numbers after the TELFIL */
      /* file name. The reason we offer a separate keyword here is to */
      /* reduce problems with exceeding CORSIKA's input line length. */
      sscanf(text+ipos,"%d %d %d %d %d",
         &max_print_tel, &max_print_evt, &skip_print, 
         &skip_print2, &skip_off2);
      if ( max_print_evt < max_print_tel )
         max_print_evt = max_print_tel;
   }
   else if ( strcasecmp(word,"internal_bunches") == 0 )
   {
      /* The maximum number of photon bunches before using external storage. */
      /* See the sixth colon separated number after the TELFIL */
      /* file name, as described for the telfil_ function. */
      if ( atoi(text+ipos) > max_internal_bunches )
         max_internal_bunches = atoi(text+ipos);
   }
   else if ( strcasecmp(word,"io_buffer") == 0 )
   {
      size_t nb=0, bs=1;
      /* The maximum size of the output buffer (usually in Megabytes). */
      /* See the seventh colon separated number after the TELFIL */
      /* file name, as described for the telfil_ function. */
      word2[0] = '\0';
      getword(text,&ipos,word2,sizeof(word2)-1,' ','\n');
      nb = atoi(word2);
      if ( strstr(word2,"M") != NULL || strstr(word2,"MB") != NULL )
         bs = 1000000;
      if ( nb < 64000 ) /* Very small numbers are implicity Megabytes. */
         bs = 1000000;
      if ( nb*bs >= 1000000 )
         max_io_buffer = nb*bs;
      setenv("CORSIKA_IO_BUFFER",word2,1);
   }
   else
   {
      fprintf(stderr,"\n\n Unknown IACT control parameter line: %s\n\n",
         word);
      exit(1); /* In the tradition of CORSIKA, everything unknown is fatal. */
   }
}

/* ---------------------------- telrne_ -------------------------- */
/**
 *  @short Write run end block to the output file.
 *
 *  @param  rune  CORSIKA run end block
 *
*/

void telrne_ (cors_real_t rune[273])
{
#ifdef HAVE_EVENTIO_FUNCTIONS
    if ( write_tel_block(iobuf,IO_TYPE_MC_RUNE,(int)rune[1],rune,3) != 0 )
       ioerrorcheck();
#endif
   printf("\n Total number of photons produced in this run: %f in %f bunches\n\n",
      all_photons_run, all_bunches_run);

}

/* --------------------------- telasu_ --------------------------- */
/**
 *  @short Setup how many times each shower is used.
 *
 *  Set up how many times the telescope system should be
 *  randomly scattered within a given area. Thus each telescope
 *  system (array) will see the same shower but at random offsets.
 *  Each shower is thus effectively used several times.
 *  This function is called according to the CSCAT keyword in the
 *  CORSIKA input file.
 *
 *  @param n   The number of telescope systems
 *  @param dx  Core range radius (if dy==0) or core x range
 *  @param dy  Core y range (non-zero for ractangular, 0 for circular)
 *  @return (none)
 *
*/

void telasu_ (int *n, cors_real_dbl_t *dx, cors_real_dbl_t *dy)
{
   core_range = core_range1 = *dx;
   core_range2 = *dy;
   nsys = *n;
}

/* --------------------------- telset_ ---------------------------- */
/**
 *  @short Add another telescope to the system (array) of telescopes.
 *
 *  Set up another telescope for the simulated telescope system.
 *  No details of a telescope need to be known except for a
 *  fiducial sphere enclosing the relevant optics.
 *  Actually, the detector could as well be a non-imaging device.
 *
 *  This function is called for each TELESCOPE keyword in the
 *  CORSIKA input file.
 *
 *  @param  x  X position [cm]
 *  @param  y  Y position [cm]
 *  @param  z  Z position [cm]
 *  @param  r  radius [cm] within which the telescope is fully contained
 *  @return (none)
 *
*/

void telset_ (cors_real_now_t *x, cors_real_now_t *y, 
   cors_real_now_t *z, cors_real_now_t *r)
{
   int itel;
   double d;
   
   if ( ((char *) x) + sizeof(cors_real_now_t) != (char *) y )
   {
      fprintf(stderr,
         "\nReal numbers in IACT code are treated as REAL*%d numbers\n",
         (int)sizeof(cors_real_now_t));
      fprintf(stderr,"but CORSIKA passes them as REAL*%d.\n",
         (int)((char *)y-(char *)x));
      exit(1);
   }
   
   if ( ntel >= MAX_ARRAY_SIZE )
   {
      fprintf(stderr,"\nToo many telescopes. Do you really have that many telescopes/detectors?\n");
      fprintf(stderr,"A little tip: increase MAX_ARRAY_SIZE in iact.c and recompile.\n");
      exit(1);
   }
   xtel[ntel] = *x;
   ytel[ntel] = *y;
   d = sqrt(xtel[ntel]*xtel[ntel]+ytel[ntel]*ytel[ntel]);
   if ( d > dmax )
      dmax = d;
   ztel[ntel] = *z + raise_tel;
   rtel[ntel] = *r;
   if ( rtel[ntel] > rmax ) 
      rmax = rtel[ntel];
   /* In order to have even the Cherenkov light from the last few meters, */
   /* all telescope z coordinates should be non-negative. */
   if ( ztel[ntel]-rtel[ntel] < 0. )
   {
      raise_tel -= ztel[ntel]-rtel[ntel];
      for (itel=0; itel<ntel; itel++)
         ztel[itel] -= ztel[ntel]-rtel[ntel];
      ztel[ntel] = rtel[ntel];
   }
   ntel++;
}

/* ------------------------- get_impact_offset ------------------------- */
/**
 *  @short Approximate impact offset of primary due to geomagnetic field.
 *
 *  Get the approximate impact offset of the primary particle due to
 *  deflection in the geomagnetic field. The approximation that the curvature
 *  radius is large compared to the distance travelled is used.
 *  The method is also not very accurate at large zenith angles where
 *  curvature of the atmosphere gets important. Therefore a zenith
 *  angle cut is applied and showers very close to zenith are skipped.
 *  Only the offset at the lowest detection level is evaluated.
 *
 *  @param  evth    CORSIKA event header block
 *  @param  prmpar  CORSIKA primary particle block. We need it to get the
 *                  particle's relativistic gamma factor (prmpar[2] or prmpar[1],
 *                  depending on the CORSIKA version).
 *  @return (none)
 *
*/

void get_impact_offset (cors_real_t evth[273], cors_real_dbl_t prmpar[PRMPAR_SIZE])
{
   int type = (int)(evth[2]+0.5);
#if CORSIKA_VERSION >= 6400
   int type2 =(int)(prmpar[0]+0.5);
#else
   int type2 =(int)(prmpar[1]+0.5);
#endif
   int curved_flag = (int)(evth[78]+0.5);

   impact_offset[0] = impact_offset[1] = 0.;

   if ( type != type2 || type <= 0 || type >= 26099 )
   {
      fprintf(stderr,"Inconsistent particle type. Skipping get_impact_offset()\n");
      return;
   }

   if ( type == 1 )  /* Gammas are not deflected */
      return;

   if ( evth[6] >= 0. )  /* With TSTART set to true, no deflection was applied in CORSIKA */
      return;

   {
      double cosz = cos(evth[10]);
      double Bxc = evth[70] * 1e-6;
      double Bzc = evth[71] * 1e-6;
      double phiB = evth[92];
      double Bx = Bxc*cos(phiB);
      double By = -Bxc*sin(phiB);
      double Bz = -Bzc;
      double theta = evth[10];
      double phi = evth[11] - evth[92];
#ifdef DEBUG_IMPACT
      double B = sqrt(Bxc*Bxc+Bzc*Bzc);
      double Bl = fabs(Bx*sin(theta)*cos(phi) + By*sin(theta)*sin(phi) + Bz*cos(theta));
      double Bt = (Bl<B?sqrt(B*B-Bl*Bl):0.);
#endif
      double charge = 0;
      double mass = 0.;
      double E = evth[3];
      double p = E;
#if CORSIKA_VERSION >= 6400
      double gamma = prmpar[1];
#else
      double gamma = prmpar[2];
#endif
      double beta = 1.;
      double c = 2.99792e8; /* m/s */
      double e = 1.602e-19; /* Coulomb */
      double psi = p*1e9*e/c;
      double dx = 0., dy = 0., dz = 0.;
      double vx = 0., vy = 0., vz = c;
#ifdef DEBUG_IMPACT
      double px = 0., py = 0., pz = p;
      double R = 1e10;
      double vx1 = 0., vy1 = 0., vz1 = c;
      double vx2 = 0., vy2 = 0., vz2 = c;
#endif
      double Fx = 0., Fy = 0., Fz = 0.;
      double t0 = fabs(evth[4]);
      double h0 = heigh_(&t0);
      double dist = (h0-obs_height)/cosz * 1e-2;
#ifdef DEBUG_IMPACT
      double t1;
#endif
      double t2;

#if CORSIKA_VERSION >= 6400
      if ( fabs(cosz-prmpar[curved_flag?15:2]) > 1e-5 )
#else
      if ( fabs(cosz-prmpar[curved_flag?15:3]) > 1e-5 )
#endif
      {
         fprintf(stderr,"Inconsistent zenith angle information: %f versus %f.\n"
            "Skipping get_impact_offset()\n",
            cosz, prmpar[curved_flag?15:((CORSIKA_VERSION>=6400)?2:3)]);
         return;
      }
      if ( cosz < 0.1 )
         return; /* Too large zenith angle or even upwards. Cannot correct. */

      if ( type >= 100 )
      {
         charge = type%100;
         mass = (0.931494-0.000511)*(int)(type/100); /* nuclear mass */
         switch ( type ) /* more accurate masses for a few nuclei */
         {
            case 402: /* 4He */
               mass = 0.931845*4.;
               break;
         }
      }
      else if ( type > 1 )
      {
         switch ( type )
         {
            case 2: /* positron */
               charge = 1;
               mass = 0.000511;
               break;
            case 5: /* mu+ */
               charge = 1;
               mass = 0.105658;
               break;
            case 8: /* pi+ */
               charge = 1;
               mass = 0.13957;
               break;
            case 11: /* K+ */
               charge = 1;
               mass = 0.493677;
               break;
            case 14: /* proton */
               charge = 1;
               mass = 0.938272;
               break;
            case 3: /* electron */
               charge = -1;
               mass = 0.000511;
               break;
            case 6: /* mu- */
               charge = -1;
               mass = 0.105658;
               break;
            case 9: /* pi- */
               charge = -1;
               mass = 0.13957;
               break;
            case 12: /* K- */
               charge = -1;
               mass = 0.493677;
               break;
            case 15: /* antiproton */
               charge = -1;
               mass = 0.938272;
               break;
            case 66: /* neutrinos */
            case 67:
            case 68:
            case 69:
               charge = 0;
               mass = 1e-8;
               break;
            default:
               fprintf(stderr,"Primary type %d not supported in get_impact_offset()\n",type);
               return;
         }
      }

#ifdef DEBUG_IMPACT
printf(" Particle type %d, mass = %f GeV/c^2, E= %g GeV, gamma = %f (%f), distance = %f km\n",
  type, mass, E, gamma, E/mass, dist*1e-3);
#endif

      if ( charge == 0 )
         return;
      if ( mass > 0. && fabs(gamma/(E/mass)-1.0) > 1e-3 )
      {
         fprintf(stderr,"Inconsistent gamma factor: %g vs. %g.\n"
            "Skipping get_impact_offset()\n", gamma, E/mass);
         return;
      }
      if ( mass > 0. && mass < E )
         p = sqrt(E*E-mass*mass); /* GeV/c */
      psi = p*1e9*e/c; /* momentum in SI units */
      if ( gamma > 1. )
         beta = sqrt(1.-1./(gamma*gamma));
      t2 = dist/(beta*c);
#ifdef DEBUG_IMPACT
      t1 = 0.5*t2;
      px = psi*sin(theta)*cos(phi); /* kg m/s */
      py = psi*sin(theta)*sin(phi);
      pz = -psi*cos(theta);
#endif
      vx = beta*c*sin(theta)*cos(phi); /* m/s */
      vy = beta*c*sin(theta)*sin(phi);
      vz = -beta*c*cos(theta);
      Fx = charge*e*(vy*Bz - vz*By);
      Fy = charge*e*(vz*Bx - vx*Bz);
      Fz = charge*e*(vx*By - vy*Bx);

#ifdef DEBUG_IMPACT
      if ( charge != 0. && Bt != 0. )
         R = p / (0.299782e2 * charge * Bt);
      printf("   R = %f m\n",R);
      printf("   p = %g kg*m/s, beta = %f, gamma = %f\n", psi, beta, gamma);
      printf("   px/p = %f, py/p = %f, pz/p = %f kg m/s\n", px/psi, py/psi, pz/psi);
      printf("   Bx = %f, By = %f, Bz = %f nT\n", Bx*1e9, By*1e9, Bz*1e9);
      printf("   Fx/p = %f, Fy/p = %f, Fz/p = %f\n", Fx/psi, Fy/psi, Fz/psi);

      printf("   Starting altitude: %f g/cm**2 -> %f km\n", evth[4], h0*1e-5);
      printf("   Ground altitude:  %f km\n", obs_height*1e-5);
      printf("   Distance to travel: %f km\n", dist*1e-3);
      printf("   Time to travel: %f s, v = %f m/s\n", t2, beta*c);

      vx1 = vx + t1*Fx/psi*(beta*c);
      vy1 = vy + t1*Fy/psi*(beta*c);
      vz1 = vz + t1*Fz/psi*(beta*c);
      vx2 = vx + t2*Fx/psi*(beta*c);
      vy2 = vy + t2*Fy/psi*(beta*c);
      vz2 = vz + t2*Fz/psi*(beta*c);

      printf("   vx0 = %f m/s, vy0 = %f m/s, vz0 = %f m/s\n", vx, vy, vz);
      printf("   vx2 = %f m/s, vy2 = %f m/s, vz2 = %f m/s\n", vx2, vy2, vz2);
      printf("   Bending of direction: %f deg\n",(180./M_PI) *
               acos((vx*vx2+vy*vy2+vz*vz2) /
                  sqrt(vx*vx+vy*vy+vz*vz) /
                  sqrt(vx2*vx2+vy2*vy2+vz2*vz2)));

      dx = (0.25*(vx+2*vx1+vx2)*t2 - vx*t2);
      dy = (0.25*(vy+2*vy1+vy2)*t2 - vy*t2);
      dz = (0.25*(vz+2*vz1+vz2)*t2 - vz*t2);
      printf("   Offsets(1): dx = %f m, dy = %f m, dz = %f m\n", dx, dy, dz);
      /* Extrapolate/interpolate from the offset point to the CORSIKA */
      /* detection level with the final velocity vector. */
      impact_offset[0] = (dx-dz*vx2/vz2) * 100.; /* in cm */
      impact_offset[1] = (dy-dz*vy2/vz2) * 100.; /* in cm */
      printf("   Impact offset(1): dxc = %f m, dyc = %f m\n", 
         impact_offset[0]*0.01, impact_offset[1]*0.01);
#endif
      
      /* These dx, dy, dz offsets are identical to the ones above. */
      dx = 0.5*t2*Fx/psi*t2*(beta*c);
      dy = 0.5*t2*Fy/psi*t2*(beta*c);
      dz = 0.5*t2*Fz/psi*t2*(beta*c);
      /* Simplified version: extrapolate to detection level with 
         initial velocity vector */
      impact_offset[0] = (dx-dz*vx/vz) * 100.; /* in cm */
      impact_offset[1] = (dy-dz*vy/vz) * 100.; /* in cm */

#ifdef DEBUG_IMPACT
      printf("   Offsets(2): dx = %f m, dy = %f m, dz = %f m\n", dx, dy, dz);
#endif

      if ( do_print )
         printf(" Impact offset: dxc = %f m, dyc = %f m\n", 
            impact_offset[0]*0.01, impact_offset[1]*0.01);
   }
}

/* --------------------------- televt_ ------------------------ */
/**
 *  @short Start of new event. Save event parameters.
 *
 *  Start of new event: get parameters from CORSIKA event header block,
 *  create randomly scattered telescope systems in given area, and
 *  write their positions as well as the CORSIKA block to the data file.
 *
 *  @param  evth    CORSIKA event header block
 *  @param  prmpar  CORSIKA primary particle block
 *  @return (none)
 *
*/

void televt_ (cors_real_t evth[273], cors_real_dbl_t prmpar[PRMPAR_SIZE])
{
#ifdef HAVE_EVENTIO_FUNCTIONS
   static int telpos_written = 0;
#endif
/*
 *    If iact.c is used but atmo.c is not, then compile iact.c with
 *    the -DNO_EXTERNAL_ATMOSPHERES compiler switch.
*/
#ifndef NO_EXTERNAL_ATMOSPHERES
   /* The model atmosphere no. is declared in atmo.c and either left zero */
   /* if the CORSIKA internal atmospheric profiles are used or set to */
   /* the user-defined atmosphere model number. */
   extern int atmosphere; 
#else
   /* No atmosphere extensions used; thus take CORSIKA atm. for granted. */
   static int atmosphere = 0;
#endif
   static int first = 1; /* For things to be written only at the start */
   int options = 0;
   int written = 0;
   int i;
   double cos_oa;


   all_photons = 0.;
   all_bunches = 0.;
   stored_bunches = 0;

   fflush(stdout);
   if ( ntel == 0 )
   {
      fprintf(stderr,
    "\n No telescopes set up. See the TELESCOPE keyword in the User's Guide.\n");
      exit(1);
   }
   event_number = (int)(evth[1]+0.5);
   /* Angles coming in radians: */
   theta_prim = evth[10];
   phi_prim = evth[11] - evth[92];
   if ( phi_prim >= 2.*M_PI )
      phi_prim -= 2.*M_PI;
   if ( phi_prim < 0.)
      phi_prim += 2.*M_PI;
#ifdef SHOW_ANGLE
   cx_prim = cos(phi_prim) * sin(theta_prim);
   cy_prim = sin(phi_prim) * sin(theta_prim);
   cz_prim = -cos(theta_prim);
#endif
   /* Angles coming in degrees: */
   theta_central = 0.5*(evth[80]+evth[81])*((double)(M_PI)/180.);
   if ( evth[83] >= evth[82] )
      phi_central = 0.5*(evth[82]+evth[83])*((double)(M_PI)/180.);
   else
   {
      phi_central = 0.5*(evth[82]+360.+evth[83])*((double)(M_PI)/180.);
   }
   phi_central -= evth[92];
   if ( phi_central >= 2.*M_PI )
      phi_central -= 2.*M_PI;
   if ( phi_central < 0.)
      phi_central += 2.*M_PI;

   ush  = sin(theta_prim)*cos(phi_prim);
   vsh  = sin(theta_prim)*sin(phi_prim);
   wsh  = cos(theta_prim);
   ushc = sin(theta_central)*cos(phi_central);
   vshc = sin(theta_central)*sin(phi_central);
   wshc = cos(theta_central);
   cos_oa = ush*ushc + vsh*vshc + wsh*wshc;
   if ( cos_oa >= 1. ) /* Catch possible rounding errors */
      off_axis = 0.;
   else
      off_axis = acos(cos_oa);

   lambda1 = evth[95];
   lambda2 = evth[96];
   energy  = evth[3];
   primary = (int) (evth[2]+0.5);
   
   { double oht = obs_height;
#ifndef NO_EXTERNAL_ATMOSPHERES
   if ( atmosphere > 0 )
      airlightspeed = 29.9792458 / refidx_(&oht);
   else
#endif
      airlightspeed = 29.9792458 / (1.+ (0.000283 * 994186.38 / 1222.656)*rhof_(&oht));
   }

   /* If the time is counted since the primary entered the atmosphere, */
   /* this is indicated by a negative number in the height of the first */
   /* interaction. */
   if ( evth[6] < 0. )
   {
      double t = 0.;
      toffset = (heigh_(&t)-obs_height) / cos(evth[10]) / 29.9792458;
      first_int = -evth[6];
   }
   else /* Time is counted since first interaction. */
   {
      toffset = (evth[6]-obs_height) / cos(evth[10]) / 29.9792458;
      first_int = evth[6];
   }

   /* Add atmospheric profile to Cherenkov flag in CORSIKA event header */
   /* if not yet present and do a few sanity checks. */
   options = (int)(evth[76]+0.5);
   if ( (options & 0x02) == 0 )
   {
      if ( first ) fprintf(stdout,
         "\n This CORSIKA version does not yet properly identify all options\n"
         " relevant for Cherenkov light production in EVTH(77).\n\n");
      options = (options & 0x3ff) | (atmosphere<=1023?atmosphere:1023 << 10);
      evth[76] = (cors_real_t) options;
   }
   else if ( (options >> 10) != atmosphere )
   {
      if ( first ) fprintf(stdout,
         "\n CORSIKA reports external atmosphere as number %d but we use %d.\n\n",
	 options >> 10, atmosphere);
      options = (options & 0x3ff) | (atmosphere<=1023?atmosphere:1023 << 10);
      evth[76] = (cors_real_t) options;
   }

   if ( (options & 0x20) != 0 )
   {
      if ( first ) fprintf(stdout,
         "\n CORSIKA was compiled with the VOLUMEDET option and the IACT option\n"
         " automatically adapts to that. This means that all random shower core\n"
         " offsets are counted in a plane perpendicular to the shower axis.\n"
         " For non-vertical showers, the horizontal offsets can therefore be\n"
         " larger than specified in your CORSIKA inputs.\n");
      options |= 0x80;
      evth[76] = (cors_real_t) options;
   }
   else
   {
      if ( first ) fprintf(stdout,
         "\n CORSIKA was compiled without the VOLUMEDET option and the IACT option\n"
         " automatically adapts to that. This means that all random shower core\n"
         " offsets are counted in a horizontal plane.\n");
   }

   if ( (options & 0x100) != 0 )
   {
      if ( first ) fprintf(stdout,
         "\n CORSIKA was compiled with the SLANT option and all longitudinal\n"
         " distributions are in slant depth units.\n");
   }
   else
   {
      if ( first ) fprintf(stdout,
         "\n CORSIKA was compiled without the SLANT option and all longitudinal\n"
         " distributions are in vertical depth units (the classical way).\n");
   }

   if ( (options & 0x04) == 0 && evth[84] < 3. )
   {
      if ( first ) fprintf(stdout,
      	 "\n You are using the IACT option (without CEFFIC) with a bunch size of %f.\n"
	 " You should be aware that this is inefficient with realistic detectors.\n"
	 " A bunch size of the order of 5 would usually be an adequate number for\n"
	 " imaging Cherenkov telescopes instrumented with ordinary photomultiplier tubes.\n",
	 evth[84]);
   }

   if ( (options & 0x04) != 0 && evth[84] > 1. )
   {
      if ( first ) fprintf(stdout,
      	 "\n You are using the IACT and CEFFIC options with a bunch size of %f.\n"
	 " Please keep in mind that this will result in correlated photo-electrons.\n",
	 evth[84]);
   }

   if ( use_compact_format ) /* We better check if it makes sense here */
   {
      int bad = 0;
      double cr = core_range;
      if ( core_range1 > cr )
         cr = core_range1;
      if ( core_range2 > cr )
         cr = core_range2;
      if ( evth[84] >= 327. )
         bad = 1;
      else if ( rmax / cos(theta_prim) > 32e2 )
         bad = 1;
      else if ( (cr+dmax+rmax) * sin(theta_prim) > 950e2 )
         bad = 1;
      if ( bad )
      {
         if ( first ) fprintf(stdout,
            "\n You selected to write output files in compact format but the configured\n"
            " values for bunch sizes, telescope sizes, positions, and random offsets\n"
            " together with the zenith angle indicate that this format is not appropriate\n"
            " and its inherent limitations would be violated.\n"
            " The compact format is disabled now.\n");
         use_compact_format = 0;
      }
   }

   first = 0;
   do_print = 0;
   if ( (nevents+1)%skip_print == 0 )
   {
      if ( count_print_evt++ < max_print_evt )
      	 do_print = 1;
   }

   if ( do_print || (nevents+1-skip_off2)%skip_print2 == 0 )
   {
      fflush(NULL); /* to get FORTRAN and C output into sync */
      fprintf(stdout,
       "\n Start with event %d (E=%5.3f TeV, first interaction in %3.1f km height)\n",
	(int)evth[1],1e-3*energy,fabs(1e-5*evth[6]));
      fprintf(stdout,
      " Azimuth=%6.2f deg (S->E) which is %6.2f deg (N->E), zenith angle=%5.2f deg\n",
           phi_prim*(180./((double)M_PI)),
           180.-phi_prim*(180./((double)M_PI)) - 
           floor((180.-phi_prim*(180./((double)M_PI)))/360.)*360.,
           theta_prim*(180./((double)M_PI)));
/*
fprintf(stdout," Reference direction is %6.2f deg (S->E), %5.2f deg zenith angle\n",
  phi_central*(180./((double)M_PI)), theta_central*(180./((double)M_PI)));
*/
      written = 1;
   }

   if ( do_print )
      fprintf(stdout," Observation level is at an altitude of %1.0f m.\n",
           0.01*obs_height);

   if ( impact_correction )
      get_impact_offset(evth,prmpar);

   set_random_systems( theta_prim, phi_prim, theta_central, phi_central,
       off_axis, energy, primary, (options&0x20));

   /* Report as many offsets as possible back to CORSIKA header block */
   for ( i=0; i<20; i++ )
      evth[98+i] = evth[118+i] = 0.;
   for ( i=0; i<20 && i<nsys; i++ )
   {
      evth[98+i]  = xoffset[i];
      evth[118+i] = yoffset[i];
   }
   evth[98-1] = i;

   televt_done = 1;
   
#ifdef HAVE_EVENTIO_FUNCTIONS
   if ( !telpos_written )
   {
      telpos_written = 1;
      if ( write_tel_pos(iobuf,ntel,xtel,ytel,ztel,rtel) != 0 ) /* 1201 */
         ioerrorcheck();
   }
   if ( write_tel_block(iobuf,IO_TYPE_MC_EVTH,(int)evth[1],evth,273) != 0 )
      ioerrorcheck();
   if ( sampling_fname != NULL ) 
   {
      /* Possibly non-uniform distribution: use weights. */
      if ( write_tel_offset_w(iobuf,nsys,toffset,xoffset,yoffset,weight) != 0 ) /* 1203 */
         ioerrorcheck();
   }
   else
   {
      /* Uniform distribution of core offsets: no weights needed. */
      if ( write_tel_offset(iobuf,nsys,toffset,xoffset,yoffset) != 0 ) /* 1203 */
         ioerrorcheck();
   }
   if ( do_print )
   {
      fprintf(stdout," Event header of event %d written to output file.\n",
      	 (int)(evth[1]+0.5));
      written = 1;
   }
   fflush(iobuf->output_file);
#endif
   if ( written )
   {
      fprintf(stdout,"\n");
      fflush(stdout);
   }
}

/* --------------------------- photon_hit ---------------------- */
/**
 *  @short Store a photon bunch for a given telescope in long format.
 *
 *  Store a photon bunch in the bunch list for a given telescope.
 *  It is kept in memory or temporary disk storage until the end
 *  of the event. This way, photon bunches or sorted by telescope.
 *  This bunch list is dynamically created and extended as required.
 *
 *  @param  det  pointer to data structure of the detector hit.
 *  @param  x    X position in CORSIKA detection plane [cm]
 *  @param  y    Y position in CORSIKA detection plane [cm]
 *  @param  cx   Direction projection onto X axis
 *  @param  cy   Direction projection onto Y axis
 *  @param  sx   Slope with respect to X axis (atan(sx) = acos(cx))
 *  @param  sy   Slope with respect to Y axis (atan(sy) = acos(cy))
 *  @param  photons Bunch size
 *  @param  ctime Arrival time of bunch in CORSIKA detection plane.
 *  @param  zem  Altitude of emission above sea level [cm]
 *  @param  lambda Wavelength (0: undetermined, -1: converted to photo-electron)
 *
 *  @return 0 (O.K.), -1 (failed to save photon bunch)
 *
*/

static int photon_hit (struct detstruct *det, double x, double y,
   double cx, double cy, double sx, double sy, double photons,
   double ctime, double zem, double lambda)
{
   struct bunch *bunch, *btmp;
   size_t wanted;
   
   det->photons += photons;

#ifdef EXTERNAL_STORAGE
   /* fprintf(stderr,"%d photon bunches so far\n",det->next_bunch);*/
   if ( det->next_bunch >= INTERNAL_LIMIT )
   {
      FILE *ext;
      sprintf(det->ext_fname,"tmp_%d_%d_%d.cbunch",
         (int)getpid(),det->iarray,det->idet);
      if ( det->external_bunches == 0 )
         unlink(det->ext_fname);
      /* fprintf(stderr,"Open file %s\n",det->ext_fname); */
#ifdef __USE_LARGEFILE64
      if ( (ext = fopen64(det->ext_fname,"a")) != NULL )
#else
      if ( (ext = fopen(det->ext_fname,"a")) != NULL )
#endif
      {
#if defined(__USE_LARGEFILE64)
         fseeko64(ext,(off64_t)sizeof(struct bunch)*
            det->external_bunches,SEEK_SET);
#elif defined(__USE_FILE_OFFSET64) || defined(__USE_LARGEFILE)
         fseeko(ext,(off_t)sizeof(struct bunch)*
            det->external_bunches,SEEK_SET);
#else
         fseek(ext,(long)sizeof(struct bunch)*
            det->external_bunches,SEEK_SET);
#endif
         if ( fwrite((void *)det->bunch,sizeof(struct bunch),
                 (size_t)det->next_bunch,ext) == det->next_bunch )
         {
            det->external_bunches += det->next_bunch;
            free(det->bunch);
            det->bunch = NULL;
            det->available_bunch = det->next_bunch = 0;
         }
         else /* if ( det->next_bunch >= 10*INTERNAL_LIMIT ) */
         {
            char cwd[1024];
            int err = errno;
            perror(det->ext_fname);
            if ( err == ENOSPC || err == EDQUOT )
               fprintf(stderr,"Too bad that you filled up your disk space or quota while storing\n"
                   "temporary photon bunches. Even though it is possible that this is just a\n"
                   "particularly large event and your final output goes to a different device,\n"
                   "you should better make sure that you have sufficient working space available\n"
                   "in directory '%s'.\n", getcwd(cwd,sizeof(cwd)));
            else
               fprintf(stderr,"That is an unexpected error while writing temporary photon bunches.\n"
                   "Check that you have sufficient space and privileges to create and write files\n"
                   "in directory '%s'.\n", getcwd(cwd,sizeof(cwd)));
            return -1;
         }
         fclose(ext);
      }
      else /* if ( det->next_bunch >= 10*INTERNAL_LIMIT ) */
      {
         char cwd[1024];
         int err = errno;
         perror(det->ext_fname);
         if ( err == ENOSPC || err == EDQUOT )
            fprintf(stderr,"Too bad that you filled up your disk space or quota while storing\n"
                "temporary photon bunches. Even though it is possible that this is just a\n"
                "particularly large event and your final output goes to a different device,\n"
                "you should better make sure that you have sufficient working space available\n"
                "in directory '%s'.\n", getcwd(cwd,sizeof(cwd)));
         else
            fprintf(stderr,"That is an unexpected error while writing temporary photon bunches.\n"
                "Check that you have sufficient space and privileges to create and write files\n"
                "in directory '%s'.\n", getcwd(cwd,sizeof(cwd)));
         return -1;
      }
   }
#endif

   if ( det->bunch == NULL )
   { 
      if ( (det->bunch = calloc(NBUNCH+EXTRA_MEM_4,sizeof(struct bunch))) == NULL )
      {
         fprintf(stderr,
    "Telescope %d of array %d: initial photon bunch list allocation failed.\n",
              det->idet, det->iarray);
         return -1;
      }
      det->available_bunch = NBUNCH;
   }

   if ( det->next_bunch >= det->available_bunch )
   {
      if ( NBUNCH < det->available_bunch/4 )
         wanted = det->available_bunch + NBUNCH + det->available_bunch/4;
      else
         wanted = det->available_bunch + NBUNCH;
      if ( (btmp = realloc(det->bunch,
             (wanted+EXTRA_MEM_4)*sizeof(struct bunch))) == NULL )
      {
         fprintf(stderr,
         "Telescope %d of array %d: photon bunch list reallocation failed.\n",
              det->idet, det->iarray);
         return -1;
      }
      det->bunch = btmp;
      det->available_bunch = wanted;
   }
   
   bunch = &det->bunch[det->next_bunch];
   bunch->photons = photons;
   bunch->x       = x - sx*det->z0 - det->x0;
   bunch->y       = y - sy*det->z0 - det->y0;
   bunch->cx      = cx;
   bunch->cy      = cy;
   bunch->ctime   = ctime - 
          det->z0*sqrt(1.+sx*sx+sy*sy)/airlightspeed - toffset;
   bunch->zem     = zem;
   bunch->lambda  = lambda;
   det->next_bunch++;
   
   return 0;
}

#if !defined(IACTEXT) && !defined(EXTENDED_TELOUT)
/* --------------------- compact_photon_hit ---------------------- */
/**
 *  @short Store a photon bunch for a given telescope in compact format.
 *
 *  Store a photon bunch in the bunch list for a given telescope.
 *  This bunch list is dynamically created and extended as required.
 *  This routine is using a more compact format than photon_hit().
 *  This compact format is not appropriate when core distances
 *  of telescopes times sine of zenith angle exceed 1000 m.
 *
 *  @param  det  pointer to data structure of the detector hit.
 *  @param  x    X position in CORSIKA detection plane [cm]
 *  @param  y    Y position in CORSIKA detection plane [cm]
 *  @param  cx   Direction projection onto X axis
 *  @param  cy   Direction projection onto Y axis
 *  @param  sx   Slope with respect to X axis (atan(sx) = acos(cx))
 *  @param  sy   Slope with respect to Y axis (atan(sy) = acos(cy))
 *  @param  photons Bunch size
 *  @param  ctime Arrival time of bunch in CORSIKA detection plane.
 *  @param  zem  Altitude of emission above sea level [cm]
 *  @param  lambda Wavelength (0: undetermined, -1: converted to photo-electron)
 *
 *  @return 0 (O.K.), -1 (failed to save photon bunch)
 *
*/

static int compact_photon_hit (struct detstruct *det, double x, double y,
   double cx, double cy, double sx, double sy, double photons,
   double ctime, double zem, double lambda)
{
   struct compact_bunch *cbunch, *cbtmp;
   size_t wanted;
   
#ifdef EXTERNAL_STORAGE
   /* fprintf(stderr,"%d photon bunches so far\n",det->next_bunch);*/
   if ( det->next_bunch >= INTERNAL_LIMIT )
   {
      FILE *ext;
      sprintf(det->ext_fname,"tmp_%d_%d_%d.cbunch",
         (int)getpid(),det->iarray,det->idet);
      if ( det->external_bunches == 0 )
         unlink(det->ext_fname);
      /* fprintf(stderr,"Open file %s\n",det->ext_fname); */
#ifdef __USE_LARGEFILE64
      if ( (ext = fopen64(det->ext_fname,"a")) != NULL )
#else
      if ( (ext = fopen(det->ext_fname,"a")) != NULL )
#endif
      {
#if defined(__USE_LARGEFILE64)
         fseeko64(ext,(off64_t)sizeof(struct compact_bunch)*
            det->external_bunches,SEEK_SET);
#elif defined(__USE_FILE_OFFSET64) || defined(__USE_LARGEFILE)
         fseeko(ext,(off_t)sizeof(struct compact_bunch)*
            det->external_bunches,SEEK_SET);
#else
         fseek(ext,(long)sizeof(struct compact_bunch)*
            det->external_bunches,SEEK_SET);
#endif
         if ( fwrite((void *)det->cbunch,sizeof(struct compact_bunch),
                 (size_t)det->next_bunch,ext) == det->next_bunch )
         {
            det->external_bunches += det->next_bunch;
            free(det->cbunch);
            det->cbunch = NULL;
            det->available_bunch = det->next_bunch = 0;
         }
         else /* if ( det->next_bunch >= 10*INTERNAL_LIMIT ) */
         {
            char cwd[1024];
            int err = errno;
            perror(det->ext_fname);
            if ( err == ENOSPC || err == EDQUOT )
               fprintf(stderr,"Too bad that you filled up your disk space or quota while storing\n"
                   "temporary photon bunches. Even though it is possible that this is just a\n"
                   "particularly large event and your final output goes to a different device,\n"
                   "you should better make sure that you have sufficient working space available\n"
                   "in directory '%s'.\n", getcwd(cwd,sizeof(cwd)));
            else
               fprintf(stderr,"That is an unexpected error while writing temporary photon bunches.\n"
                   "Check that you have sufficient space and privileges to create and write files\n"
                   "in directory '%s'.\n", getcwd(cwd,sizeof(cwd)));
            return -1;
         }
         fclose(ext);
      }
      else /* if ( det->next_bunch >= 10*INTERNAL_LIMIT ) */
      {
         char cwd[1024];
         int err = errno;
         perror(det->ext_fname);
         if ( err == ENOSPC || err == EDQUOT )
            fprintf(stderr,"Too bad that you filled up your disk space or quota while storing\n"
                "temporary photon bunches. Even though it is possible that this is just a\n"
                "particularly large event and your final output goes to a different device,\n"
                "you should better make sure that you have sufficient working space available\n"
                "in directory '%s'.\n", getcwd(cwd,sizeof(cwd)));
         else
            fprintf(stderr,"That is an unexpected error while writing temporary photon bunches.\n"
                "Check that you have sufficient space and privileges to create and write files\n"
                "in directory '%s'.\n", getcwd(cwd,sizeof(cwd)));
         return -1;
      }
   }
#endif

   det->photons += photons;

   if ( det->cbunch == NULL )
   { 
      if ( (det->cbunch = calloc(NBUNCH+EXTRA_MEM_4,sizeof(struct compact_bunch))) == NULL )
      {
         fprintf(stderr,
    "Telescope %d of array %d: initial photon bunch list allocation failed.\n",
              det->idet, det->iarray);
         return -1;
      }
      det->available_bunch = NBUNCH;
   }
   
   if ( det->next_bunch >= det->available_bunch )
   {
      if ( NBUNCH < det->available_bunch/4 )
         wanted = det->available_bunch + NBUNCH + det->available_bunch/4;
      else
         wanted = det->available_bunch + NBUNCH;
      if ( (cbtmp = realloc(det->cbunch,
             (wanted+EXTRA_MEM_4)*sizeof(struct compact_bunch))) == NULL )
      {
         fprintf(stderr,
         "Telescope %d of array %d: photon bunch list reallocation failed.\n",
              det->idet, det->iarray);
         return -1;
      }
      det->cbunch = cbtmp;
      det->available_bunch = wanted;
   }
   
   cbunch = &det->cbunch[det->next_bunch];
   /*@ The bunch size has a limited range of up to 327 photons. */
   cbunch->photons = (short)(100.*photons+0.5);
   /*@ Positions have a limited range of up to 32.7 m from the detector centre */
   cbunch->x       = (short)Nint(10.*(x - sx*det->z0 - det->x0));
   cbunch->y       = (short)Nint(10.*(y - sy*det->z0 - det->y0));
   /*@ No limits in the direction (accuracy 7 arcsec for vertical showers) */
   cbunch->cx      = (short)Nint(30000.*cx);
   cbunch->cy      = (short)Nint(30000.*cy);
   /*@ The time has a limited range of +-3.27 microseconds. */
   cbunch->ctime   = (short)Nint(10.*(ctime - 
           det->z0*sqrt(1.+sx*sx+sy*sy)/airlightspeed - toffset));
   cbunch->log_zem = (short)(1000.*log10(zem)+0.5);
   if ( lambda == 0. )     /* Unspecified wavelength of photons */
      cbunch->lambda = (short)0;
   else if ( lambda < 0. ) /* Photo-electrons instead of photons */
      cbunch->lambda = (short)(lambda-0.5);
   else                    /* Photons of specific wavelength */
      cbunch->lambda = (short)(lambda+0.5);
   det->next_bunch++;
   
   return 0;
}
#endif

/* -------------------------- Nint_f ------------------------- */
/**
 *
 *  Nearest integer function
 *
*/

static int Nint_f(double x)
{
   if ( x >= 0. )
      return (int)(x+0.5);
   else
      return (int)(x-0.5);
}


/* --------------------------- telout_ --------------------------- */
/**
 *  @short Check if a photon bunch hits one or more simulated detector volumes.

 *  A bunch of photons from CORSIKA is checked if they hit a
 *  a telescope and in this case it is stored (in memory).
 *  This routine can alternatively trigger that the photon bunch
 *  is written by CORSIKA in its usual photons file.
 *
 *  Note that this function should only be called for downward photons
 *  as there is no parameter that could indicate upwards photons.
 *
 *  The interface to this function can be modified by defining
 *  EXTENDED_TELOUT. Doing so requires to have a CORSIKA version
 *  with support for the IACTEXT option, and to actually activate
 *  that option. That could be useful when adding your own
 *  code to create some nice graphs or statistics that requires
 *  to know the emitting particle and its energy but would be of
 *  little help for normal use. Inconsistent usage of 
 *  EXTENDED_TELOUT here and IACTEXT in CORSIKA will most likely
 *  lead to a crash.

 *  @param  bsize   Number of photons (can be fraction of one)
 *  @param  wt	   Weight (if thinning option is active)
 *  @param  px	   x position in detection level plane
 *  @param  py	   y position in detection level plane
 *  @param  pu	   x direction cosine
 *  @param  pv	   y direction cosine
 *  @param  ctime   arrival time in plane after first interaction
 *  @param  zem     height of emission above sea level
 *  @param  lambda  0. (if wavelength undetermined) or wavelength [nm].
 *                  If lambda < 0, photons are already converted to
 *                  photo-electrons (p.e.), i.e. we have p.e. bunches.
 *  @param  temis   Time of photon emission (only if CORSIKA extracted
 *                  with IACTEXT option and this code compiled with
 *                  EXTENDED_TELOUT defined).
 *  @param  penergy Energy of emitting particle (under conditions as temis).
 *  @param  amass   Mass of emitting particle (under conditions as temis).
 *  @param  charge  Charge of emitting particle (under conditions as temis).
 *
 *  @return  0 (no output to old-style CORSIKA file needed)
 *           2 (detector hit but no eventio interface available or
 *             output should go to CORSIKA file anyway)
 *
*/

int telout_ (cors_real_now_t *bsize, cors_real_now_t *wt, 
   cors_real_now_t *px, cors_real_now_t *py, 
   cors_real_now_t *pu, cors_real_now_t *pv, 
   cors_real_now_t *ctime, cors_real_now_t *zem , cors_real_now_t *lambda
#ifdef EXTENDED_TELOUT   
   , double *temis, double *penergy, double *amass, double *charge
#endif
   )
{
   int ix, iy, igrid, rc, kdet, idet, iarray;
   double x = *px - impact_offset[0], y = *py - impact_offset[1];
   double u = *pu, v = *pv, sx, sy;
   double xphotons = *bsize;
   double wl = *lambda;
#ifdef CORSIKA_SAVES_PHOTONS
   int is_corsika;
#endif

   all_photons += xphotons;
   all_bunches += 1.0;   

   /* Completely outside of grid area? */
   if ( x < grid_x_low || y < grid_y_low )
      return 0;
   ix = (int)((x-grid_x_low)/GRID_SIZE);
   iy = (int)((y-grid_y_low)/GRID_SIZE);
   /* Completely outside of grid area? */
   if ( ix < 0 || ix >= grid_nx || iy < 0 || iy >= grid_ny )
      return 0;

   igrid = iy*grid_nx + ix;
   if ( iy*grid_nx+ix < 0 || iy*grid_nx+ix >= grid_elements )
   {
      fprintf(stderr,
     "Grid bounds exceeded***: ix=%d, iy=%d, i=%d (nx=%d, ny=%d, n=%d)\n",
         ix, iy, iy*grid_nx+ix, grid_nx, grid_ny, grid_elements);
      return -1;
   }
   /* No telescopes/detectors on this grid element? */
   if ( grid[igrid].ndet <= 0 )
      return 0;

#ifdef CORSIKA_SAVES_PHOTONS
   is_corsika = 0;
#endif

   sx = u/sqrt(1.-u*u-v*v);
   sy = v/sqrt(1.-u*u-v*v);


   /* Check all telescopes/detectors having something in this */
   /* grid element if the photons hit this telescope/detector. */
   for (kdet=0; kdet<grid[igrid].ndet; kdet++)
      /* First a quick check if the photons hit the projection of the */
      /* somewhat enlarged detector circumference on the detection level. */
      if ( fabs(x-grid[igrid].detectors[kdet]->x) <= 
              grid[igrid].detectors[kdet]->r )
       if ( fabs(y-grid[igrid].detectors[kdet]->y) <= 
               grid[igrid].detectors[kdet]->r )
        /* Then do a more careful check */
        if ( (rc = in_detector(grid[igrid].detectors[kdet],x,y,sx,sy)) != 0 )
        {
           /* Yes the photons really hit */
#ifdef EXTENDED_TELOUT
           if ( photon_hit(grid[igrid].detectors[kdet],
                     x, y, u, v, sx, sy, xphotons,
                    (double)*ctime, (double)*zem, wl) != 0 
                || photon_hit(grid[igrid].detectors[kdet],
                     x, y, *amass, *charge, sx, sy, *penergy,
                     (double)*ctime, *temis, 9999) != 0 )
#else
           if ( (use_compact_format ?
                  compact_photon_hit(grid[igrid].detectors[kdet],
                     x, y, u, v, sx, sy, xphotons,
                    (double)*ctime, (double)*zem, wl) :
                  photon_hit(grid[igrid].detectors[kdet],
                     x, y, u, v, sx, sy, xphotons,
                    (double)*ctime, (double)*zem, wl)) != 0 )
#endif
           {
              fflush(NULL);
              fprintf(stderr,"\n******************************************\n");
              fprintf(stderr,"Fatal problem with storing photon bunches.\n");
              fprintf(stderr,"Already stored: %ld bunches.\n",stored_bunches);
              fprintf(stderr,"Run this simulation with a smaller energy\n");
              fprintf(stderr,"or with a smaller number of telescopes or\n");
              fprintf(stderr,"run it on a computer with more memory.\n");
              fprintf(stderr,"Make also sure you have enough disk space.\n");
              fprintf(stderr,"******************************************\n\n");
              for (iarray=0; iarray<narray; iarray++)
                 for (idet=0; idet<ndet[iarray]; idet++)
                    if ( detector[iarray][idet].ext_fname[0] != '\0' )
                       unlink(detector[iarray][idet].ext_fname);
              exit(1);
           }
           stored_bunches++;
#ifdef CORSIKA_SAVES_PHOTONS
           is_corsika = 2;
#endif
        }           

#ifndef CORSIKA_SAVES_PHOTONS
   return 0;
#else
   return is_corsika;
#endif
}

#ifdef IACTEXT
/* --------------------------- telprt_ ------------------------- */
/**
 *  @short Store CORSIKA particle information into IACT output file.
 *
 *  This function is not needed for normal simulations and is
 *  therefore only available if the preprocessor symbols
 *  IACTEXT or EXTENDED_TELOUT are defined. At the same time
 *  CORSIKA itself should be extracted with the IACTEXT option.
 *
 *  @param datab  A particle data buffer with up to 39 particles.
 *  @param maxbuf The buffer size, which is 39*7 without thinning
 *                option and 39*8 with thinning.
 */

void telprt_ (cors_real_t *datab, int *maxbuf)
{
   int i;
   int vpp = (*maxbuf)/39; /* Values per particle (7 without thinning, 8 with) */
   int thinning = (vpp==8?1:0);

   if ( particles == NULL )
   {
      if ( (particles = calloc(NBUNCH+EXTRA_MEM_4,sizeof(struct bunch))) == NULL )
      {
         fprintf(stderr, "Initial particles list allocation failed.\n");
         return;
      }
      particles_space = NBUNCH;
      particles_stored = 0;
   }

   if ( particles_stored+39 > particles_space )
   {
      size_t wanted = particles_space + NBUNCH;
      struct bunch *tmpp = realloc(particles, 
            (wanted+EXTRA_MEM_4)*sizeof(struct bunch));
      if ( tmpp == NULL )
      {
         fprintf(stderr,"Particle list reallocation failed.\n");
         return;
      }
      particles = tmpp;
      particles_space = wanted;
   }

   for (i=0; i<39; i++ )
   {
      if ( datab[i*vpp] == 0. ) /* Buffer not filled completely */
         break;

      double p = sqrt(datab[i*vpp+1]*datab[i*vpp+1] +
          datab[i*vpp+2]*datab[i*vpp+2] +
          datab[i*vpp+3]*datab[i*vpp+3]);
      struct bunch *bunch = &particles[particles_stored];
      bunch->zem    = datab[i*vpp+3]>0. ? p : -p; /* momentum goes into zem; check sign */
      bunch->x      = datab[i*vpp+4] - impact_offset[0];
      bunch->y      = datab[i*vpp+5] - impact_offset[1];
      bunch->cx     = datab[i*vpp+1] / p;
      bunch->cy     = datab[i*vpp+2] / p;
      bunch->ctime  = datab[i*vpp+6];
      bunch->lambda = datab[i*vpp]; /* Particle information goes into lamda */
      if ( thinning )
         bunch->photons = datab[i*vpp+7]; /* Thinning weight */
      else
         bunch->photons = 1.;
      particles_stored++;

#ifdef DEBUG_IMPACT
      printf(" Particle type %d hits detection level %d at x=%f m, y=%f m\n",
         ((int)(bunch->lambda+0.5))/1000,
         ((int)(bunch->lambda+0.5))%1000,
         bunch->x*1e-2,bunch->y*1e-2 );
      printf("   corrected for offset %f m, %f m\n", 
         impact_offset[0]*1e-2, impact_offset[1]*1e-2);
#endif
   }

   return;
}
#endif

/* --------------------------- tellng_ ------------------------- */
/**
 *  @short Write CORSIKA 'longitudinal' (vertical) distributions.
 *
 *  Write several kinds of vertical distributions to the output.
 *  These or kinds of histograms as a function of atmospheric depth.
 *  In CORSIKA, these are generally referred to as 'longitudinal' distributions.
 *
 *  @verbatim
 *  There are three types of distributions:
 *	type 1: particle distributions for
 *		gammas, positrons, electrons, mu+, mu-,
 *		hadrons, all charged, nuclei, Cherenkov photons.
 *	type 2: energy distributions (with energies in GeV) for
 *		gammas, positrons, electrons, mu+, mu-,
 *		hadrons, all charged, nuclei, sum of all.
 *	type 3: energy deposits (in GeV) for
 *		gammas, e.m. ionisation, cut of e.m.  particles,
 *		muon ionisation, muon cut, hadron ionisation,
 *		hadron cut, neutrinos, sum of all.
 *		('cut' accounting for low-energy particles dropped)
 *  @endverbatim
 *
 *  Note: Corsika can be extracted from CMZ sources with three options
 *  concerning the vertical profile of Cherenkov light:
 *  default = emission profile, INTCLONG = integrated light profile,
 *  NOCLONG = no Cherenkov profiles at all. If you know which kind
 *  you are using, you are best off by defining it for compilation
 *  of this file (either -DINTEGRATED_LONG_DIST, -DEMISSION_LONG_DIST, or
 *  -DNO_LONG_DIST).
 *  By default, a run-time detection is attempted which should work well
 *  with some 99.99% of all air showers but may fail in some cases like 
 *  non-interacting muons as primary particles etc.
 *
 *  @param  type    see above
 *  @param  data    set of (usually 9) distributions
 *  @param  ndim    maximum number of entries per distribution
 *  @param  np      number of distributions (usually 9)
 *  @param  nthick  number of entries actually filled per distribution
 *                  (is 1 if called without LONGI being enabled).
 *  @param  thickstep  step size in g/cm**2
 *
 *  @return  (none)
 *
*/

void tellng_ (int *type, double *data, int *ndim, int *np, 
   int *nthick, double *thickstep)
{
#if defined(INTEGRATED_LONG_DIST)
   static int is_integ = 1;
#elif defined(UNKNOWN_LONG_DIST)
   static int is_integ = 0;
#else
   static int is_integ = -1;
#endif
   static int init_done = 0;
   int i;

#ifdef HAVE_EVENTIO_FUNCTIONS
#ifndef WRITE_ALL_LONG
   /* Ignoring everything except particle distributions */
   if ( *type != 1 )
      return;
#endif

   if ( *nthick <= 1 )
   {
      if ( !init_done )
      	 printf("\n No vertical profiles switched on and none written to output.\n\n");
      init_done = 1;
      return;
   }

#ifdef NO_LONG_DIST
   if ( !init_done )
      printf("\n No vertical profiles are written to output.\n\n");
   init_done = 1;
   return;
#else
#ifdef UNKNOWN_LONG_DIST
   /* When not knowing how CORSIKA writes Cherenkov light longitudinal */
   /* distributions, we have to find it out dynamically. */
   if ( *type == 1 && is_integ == 0 )
   {
      double sum_prof = data[8*(*ndim)+0];
      /* If the photon number ever falls of again it is an emission profile. */
      for (i=1; i<(*nthick); i++)
      {
      	 sum_prof += data[8*(*ndim)+i];
         if ( data[8*(*ndim)+i] - data[8*(*ndim)+i-1] < 0 )
      	 {
	    printf("\n Cherenkov light vertical distribution is an"
	           "\n emission profile and will not be differentiated.\n\n");
	    is_integ = -1;
	    break;
	 }
      }
      if ( is_integ == 0 )
      {
      	 if ( sum_prof == 0. )
	    printf("\n Cherenkov light vertical distribution is empty.\n\n");
	 else
            printf("\n Cherenkov light vertical distribution is presumably"
	        "\n an integrated profile and will be differentiated.\n\n");
      }
      /* Note that in the latter case there is a small chance of being wrong. */
   }
#endif

   /* Cherenkov distribution is cummulative: differentiate it now */
   if ( *type == 1 && is_integ != -1 )
   {
      for (i=(*nthick)-1; i>=1; i--)
      	 data[8*(*ndim)+i] -= data[8*(*ndim)+i-1];
   }

#ifdef DEBUG_LONG_DIST
   if ( *type == 1 )
   {
      for (i=0; i<(*nthick); i++)
      {
	 double heigh_(double *x);
	 double x=(*thickstep)*(i+0.5);
	 printf("# %7.3f %f %g %g\n",heigh_(&x)*1e-5,x/cos(theta_prim),
	    data[6*(*ndim)+i],data[8*(*ndim)+i]);
      }
   }
#endif

   /* Now actually store the longitudinal distributions */
   if ( write_shower_longitudinal(iobuf,event_number,*type,data,*ndim,*np,
            *nthick,*thickstep) != 0 )
      ioerrorcheck();

   /* Restore cumulative Cherenkov distribution */
   if ( *type == 1 && is_integ != -1 )
   {
      for (i=1; i<(*nthick); i++)
      	 data[8*(*ndim)+i] += data[8*(*ndim)+i-1];
   }
#endif /* NO_LONG_DIST */
#endif /* HAVE_EVENTIO_FUNCTIONS */
}

/* --------------------------- telend_ ------------------------- */
/**
 *  @short End of event. Write out all recorded photon bunches.
 *
 *  End of an event: write all stored photon bunches to the
 *  output data file, and the CORSIKA event end block as well.
 *
 *  @param  evte  CORSIKA event end block
 *  @return (none)
 *
*/

void telend_ (cors_real_t evte[273])
{
   int iarray, idet;
#ifdef HAVE_EVENTIO_FUNCTIONS
   IO_ITEM_HEADER item_header;
#endif
   
   do_print = 0;
   if ( (nevents+1)%skip_print == 0 )
   {
      if ( count_print_tel++ < max_print_tel )
      	 do_print = 1;
   }

   all_photons_run += all_photons;
   all_bunches_run += all_bunches;
   nevents++;

   if ( do_print )
   {
      printf("\n Total number of photons in shower: %f in %f bunches\n",
         all_photons,all_bunches);

      for (iarray=0; iarray<narray; iarray++)
      {
	 printf(" Array %2d: ",iarray);
	 for (idet=0; idet<ndet[iarray]; idet++)
	 {
            printf(" %6.0f",detector[iarray][idet].photons);
	 }
	 printf(" photons (array offset: %7.1f %7.1f m)\n",
            xoffset[iarray]*0.01,yoffset[iarray]*0.01);
      }
      fflush(NULL);
   }

#ifdef IACTEXT
   /* If we have extra particle information, write that first. */
   if ( particles != NULL && particles_stored > 0 )
   {
      if ( write_tel_photons(iobuf, 999, 999, (double) particles_stored,
             particles, particles_stored, 0, NULL) != 0 )
         ioerrorcheck();
      particles_stored = 0;
      if ( particles != NULL )
      {
         free(particles);
         particles = NULL;
         particles_space = 0;
      }
   }
#endif

#ifdef HAVE_EVENTIO_FUNCTIONS
   for (iarray=0; iarray<narray; iarray++)
   {
      begin_write_tel_array(iobuf,&item_header,iarray);
      for (idet=0; idet<ndet[iarray]; idet++)
      {
#ifdef WRITE_ONLY_NONEMPTY_TELESCOPES
         if ( detector[iarray][idet].next_bunch + 
              detector[iarray][idet].external_bunches > 0 )
#endif
         {
            if ( use_compact_format )
            {
               if ( write_tel_compact_photons(iobuf,iarray,idet,
                     detector[iarray][idet].photons,
                     detector[iarray][idet].cbunch,
                     detector[iarray][idet].next_bunch,
                     detector[iarray][idet].external_bunches,
                     detector[iarray][idet].ext_fname) != 0 )
                  ioerrorcheck();
            }
            else
            {
               if ( write_tel_photons(iobuf,iarray,idet,
                     detector[iarray][idet].photons,
                     detector[iarray][idet].bunch,
                     detector[iarray][idet].next_bunch,
                     detector[iarray][idet].external_bunches,
                     detector[iarray][idet].ext_fname) != 0 )
                  ioerrorcheck();
            }
         }
         if ( detector[iarray][idet].ext_fname[0] != '\0' )
            unlink(detector[iarray][idet].ext_fname);
         detector[iarray][idet].external_bunches = 0;
         detector[iarray][idet].ext_fname[0] = '\0';
      }
      if ( end_write_tel_array(iobuf,&item_header) != 0 )
         ioerrorcheck();
   }
   if ( write_tel_block(iobuf,IO_TYPE_MC_EVTE,(int)evte[1],evte,273) != 0 )
      ioerrorcheck();
   fflush(iobuf->output_file);
#endif

}

/* ----------------------------------------------------------- */

#if 0
/* in case of problems with square() implemented as a macro use function */
static double square(double x)
{
   return x*x;
}
#else
#define square(x) ((x)*(x))
#endif

/* ------------------------- in_detector --------------------- */
/**
 *  @short Check if a photon bunch hits a particular telescope volume.
 *
 *  Check if a photon bunch (or, similarly, a particle) hits a
 *  particular simulated telescope/detector.
 *
 *  @param  x  X position of photon position in CORSIKA detection level [cm]
 *  @param  y  Y position of photon position in CORSIKA detection level [cm]
 *  @param  sx Slope of photon direction in X/Z plane.
 *  @param  sy Slope of photon direction in Y/Z plane.
 *
 *  @return 0 (does not hit), 1 (does hit)
*/

static int in_detector (det, x, y, sx, sy)
  struct detstruct *det;
  double x, y, sx, sy;
{
   int rc;
   double xd, yd, d2;
   rc = 0;

   switch ( det->geo_type )
   {
      case 1: /* circular (in detection level plane) */
         if ( sqrt((x-det->x)*(x-det->x)+(y-det->y)*(y-det->y)) > det->dx )
            return 0;
         break;
      case 2: /* rectangular (in detection level plane) */
         if ( fabs(x-det->x) > det->dx || fabs(y-det->y) > det->dy )
            return 0;
         break;
      case 3: /* quasi spherical (fiducial volume in space) */
         xd = x - sx*det->z0;
         yd = y - sy*det->z0;
         
#ifdef DEBUG_IACT
printf("Photon: x=%5.0f, y=%5.0f, sx=%5.3f, sy=%5.3f, xd=%5.0f, yd=%5.0f\n",
      x,y,sx,sy,xd,yd);
printf("Detektor: x=%5.0f, y=%5.0f, z=%3.0f\n",
      det->x0,det->y0,det->z0);
fflush(stdout);
#endif

         /* Distance of photon path at (xd,yd,det->z0) with direction */
         /* (sx,sy,1) from centre of sphere at (det->x0,det->y0,det->z0). */
         d2 = (square((xd-det->x0)*sy-(yd-det->y0)*sx) +
               square((yd-det->y0)) +
               square((xd-det->x0))) / (sx*sx+sy*sy+1.);
         /* Reject paths not intersecting sphere. */
         /* Note: photons are accepted even if produced below the sphere. */
         if ( d2 > det->r0*det->r0 )
            return 0;
#ifdef DEBUG_IACT
printf("Telescope hit, d = %5.0f\n",sqrt(d2));
fflush(stdout);
#endif
         break;
   }

   return 1;
}

/* ------------------------ set_random_systems ---------------------- */
/**
 *  @short Randomly scatter each array of detectors in given area.
 *
 *  The area containing the detectors is sub-divided into a rectangular
 *  grid and each detector with a (potential) intersection with a grid
 *  element is marked for that grid element. A detector can be marked
 *  for several grid elements unless completely inside one element.
 *  Checks which detector(s) is/are hit by a photon bunch (or, similarly,
 *  by a particle) is thus reduced to check only the detectors marked
 *  for the grid element which is hit by the photon bunch (or particle).
 *  The grid should be sufficiently fine-grained that there are usually
 *  not much more than one detector per element but finer graining than
 *  the detector sizes makes no sense.
 *
 *  @param theta     Zenith angle of the shower following [radians].
 *  @param phi       Shower azimuth angle in CORSIKA angle convention [radians].
 *  @param thetaref  Reference zenith angle (e.g. of VIEWCONE centre) [radians].
 *  @param phiref    Reference azimuth angle (e.g. of VIEWCONE centre) [radians].
 *  @param offax     Angle between central direction (typically VIEWCONE centre)
 *                   and the direction of the current primary [radians].
 *  @param E         Primary particle energy in GeV (may be used in importance sampling).
 *  @param primary   Primary particle ID (may be used in importance sampling).
 *  @param volflag   Set to 1 if CORSIKA was compiled with VOLUMEDET option, 0 otherwise.
 *
 *  @return 0 (O.K.), -1 (error)
*/

static int set_random_systems (double theta, double phi, 
   double thetaref, double phiref, double offax,
   double E, int primary, int volflag)
{
   int iarray, idet, iclass, ix, iy; 
   int start_class[MAX_CLASS], size_class[MAX_CLASS];
   double r_class[MAX_CLASS], dx_class[MAX_CLASS], dy_class[MAX_CLASS];
   int nod, nn;

   narray = nsys;
   if ( ndet == NULL )
      ndet = (int *) calloc((size_t)narray+EXTRA_MEM_5,sizeof(int));

   size_class[0] = nsys;
   for (start_class[0]=0,iclass=1; iclass<MAX_CLASS; iclass++)
      start_class[iclass] = start_class[iclass-1]+size_class[iclass-1];

   for (iarray=start_class[0]; iarray<start_class[0]+size_class[0]; iarray++)
      ndet[iarray] = ntel;
      
   for (iclass=0; iclass<MAX_CLASS; iclass++)
     det_in_class[iclass] = ndet[start_class[iclass]];

   for (iarray=nod=0; iarray<narray; iarray++)
      nod += ndet[iarray];
   if ( do_print )
      fprintf(stdout," %d telescope%s simulated in %d array%s.\n",
         nod, nod==1?" is":"s are", narray, narray==1?"":"s");

   if ( detector == NULL )
      detector = (struct detstruct **) calloc((size_t)narray+EXTRA_MEM_6,
          sizeof(struct detstruct *));
   if ( xoffset == NULL )
      xoffset = calloc((size_t)narray+EXTRA_MEM_7,sizeof(double));
   if ( yoffset == NULL )
      yoffset = calloc((size_t)narray+EXTRA_MEM_7,sizeof(double));
   if ( weight == NULL )
      weight = calloc((size_t)narray+EXTRA_MEM_7,sizeof(double));

   for (iarray=0; iarray<narray; iarray++)
   {
      if ( detector[iarray] == NULL )
       if ( (detector[iarray] = 
           calloc((size_t)ndet[iarray]+EXTRA_MEM_8,sizeof(struct detstruct))) == NULL )
      {
         fprintf(stderr,"Detector allocation failed\n");
         return -1;
      }
   }

   iclass = 0;
   for (iarray=start_class[iclass]; 
        iarray<start_class[iclass]+size_class[iclass]; iarray++)
   {
      for ( idet = 0; idet < ndet[iarray]; idet++ )
      {
         /* Projected position (along shower direction) on detection level: */
         detector[iarray][idet].x = xtel[idet] + ztel[idet]*tan(theta)*cos(phi);
         detector[iarray][idet].y = ytel[idet] + ztel[idet]*tan(theta)*sin(phi);
         /* Simple fiducial area is larger than simple projection: */
         detector[iarray][idet].r  = 
         detector[iarray][idet].dx = 
         detector[iarray][idet].dy = 
              rtel[idet]/cos(theta) * 1.1 + fabs(ztel[idet]*tan(theta)*0.1);
         /* Position in space: */
         detector[iarray][idet].x0 = xtel[idet];
         detector[iarray][idet].y0 = ytel[idet];
         detector[iarray][idet].z0 = ztel[idet];
         detector[iarray][idet].r0 = rtel[idet];
         detector[iarray][idet].geo_type = 3;
         detector[iarray][idet].sens_type = 1;
         detector[iarray][idet].class = iclass;
         if ( detector[iarray][idet].bunch != NULL )
            free(detector[iarray][idet].bunch);
         detector[iarray][idet].bunch = NULL;
         if ( detector[iarray][idet].cbunch != NULL )
            free(detector[iarray][idet].cbunch);
         detector[iarray][idet].cbunch = NULL;
         detector[iarray][idet].available_bunch = 0;
         detector[iarray][idet].next_bunch = 0;
         detector[iarray][idet].photons = 0.;
         detector[iarray][idet].ext_fname[0] = '\0';
         detector[iarray][idet].external_bunches = '\0';
      }
   }

   /* Determine size of arrays. */

   for ( iclass=0; iclass<MAX_CLASS; iclass++ )
   {
      r_class[iclass] = dx_class[iclass] = dy_class[iclass] = 0.;
      iarray = start_class[iclass];
      for ( idet=0; idet<ndet[iarray]; idet++ )
      {
         double r, dx, dy;

         r = sqrt(detector[iarray][idet].x*detector[iarray][idet].x +
                  detector[iarray][idet].y*detector[iarray][idet].y) +
             detector[iarray][idet].r;
         dx = fabs(detector[iarray][idet].x)+detector[iarray][idet].dx;
         dy = fabs(detector[iarray][idet].y)+detector[iarray][idet].dy;
         if ( r > r_class[iclass] )
            r_class[iclass] = r;
         if ( dx > dx_class[iclass] )
            dx_class[iclass] = dx;
         if ( dy > dy_class[iclass] )
            dy_class[iclass] = dy;
      }
   }

   /* Shift detectors by random amount (but must be inside trigger area) */

   for ( iarray = 0; iarray < narray; iarray++ )
   {
      int inside;
      double xoff, yoff;

      iclass = detector[iarray][0].class;
      
      if ( sampling_fname == NULL )
      {
         /* Uniform sampling */
         for ( inside=0; !inside; )
         {
            if ( core_range2 <= 0. )
            {
               xoff = core_range*(2.*(rndm(0)-0.5));
               yoff = core_range*(2.*(rndm(1)-0.5));
               if ( sqrt(xoff*xoff+yoff*yoff) <= core_range )
                  inside = 1;
            }
            else
            {
               xoff = core_range1*(2.*(rndm(0)-0.5));
               yoff = core_range2*(2.*(rndm(1)-0.5));
               inside = 1;
            }

#ifdef DEBUG_IACT
printf("Array %d (class%d) has offset dx=%f, dy=%f\n",iarray,iclass,xoff,yoff);
fflush(stdout);
#endif
         }

         if ( core_range2 <= 0. )
            detector[iarray][0].sampling_area = M_PI*(core_range*core_range) / (double)narray;
         else
            detector[iarray][0].sampling_area = 4.*(core_range1*core_range2) / (double)narray;
      }
      else if ( core_range2 != 0. )
      {
         fprintf(stderr,"\n Importance sampling can only be used with circular regions.\n");
         exit(1);
      }
      else if ( !volflag )
      {
         fprintf(stderr,"\n Importance sampling with circular regions requires CORSIKA option VOLUMEDET.\n");
         exit(1);
      }
      else
         sample_offset(sampling_fname, core_range, theta, phi, 
            thetaref, phiref, offax, E, primary,
            &xoff, &yoff, &detector[iarray][0].sampling_area);

      /* If CORSIKA is compiled with the VOLUMEDET option, we follow that */
      /* here and take our random offset to be in a plane perpendicular to */
      /* the shower axis and not a horizontal plane. */

      if ( volflag )
      {
         double x1, y1, x2, y2;
         x1 = xoff/cos(theta);
         y1 = yoff;
         x2 = x1*cos(phi) - y1*sin(phi);
         y2 = x1*sin(phi) + y1*cos(phi);
         xoff = x2;
         yoff = y2;
      }

      xoffset[iarray] = xoff;
      yoffset[iarray] = yoff;
      weight[iarray] = detector[iarray][0].sampling_area;
      for ( idet = 0; idet < ndet[iarray]; idet++ )
      {
         detector[iarray][idet].x += xoff;
         detector[iarray][idet].y += yoff;
         detector[iarray][idet].x0 += xoff;
         detector[iarray][idet].y0 += yoff;
      }
#ifdef DEBUG_IACT
printf("Array %d (class%d) has offset dx=%f, dy=%f\n",iarray,iclass,xoff,yoff);
fflush(stdout);
#endif
   }

   /* Free grid from previous showers */
   
   if ( grid != NULL )
   {
      for (iy=0; iy<grid_ny; iy++)
         for (ix=0; ix<grid_nx; ix++)
         {
            struct gridstruct *gc;
            if ( iy*grid_nx+ix < 0 || iy*grid_nx+ix >= grid_elements )
            {
               fprintf(stderr,
              "Grid bounds exceeded: ix=%d, iy=%d, i=%d (nx=%d, ny=%d, n=%d)\n",
                  ix, iy, iy*grid_nx+ix, grid_nx, grid_ny, grid_elements);
               return -1;
            }
            gc = &grid[iy*grid_nx+ix];
            if ( gc->detectors != NULL )
            {
               free(gc->detectors);
               gc->detectors = NULL;
            }
            gc->idet = gc->ndet = 0;
         }
      free(grid);
      grid = NULL;
   }

   /* Determine grid size for next shower */

   for (iarray=0; iarray < narray; iarray++)
      for (idet=0; idet<ndet[iarray]; idet++)
      {
         if ( detector[iarray][idet].x - detector[iarray][idet].r < grid_x_low )
            grid_x_low = detector[iarray][idet].x - detector[iarray][idet].r;
         if ( detector[iarray][idet].x + detector[iarray][idet].r > grid_x_high )
            grid_x_high = detector[iarray][idet].x + detector[iarray][idet].r;
         if ( detector[iarray][idet].y - detector[iarray][idet].r < grid_y_low )
            grid_y_low = detector[iarray][idet].y - detector[iarray][idet].r;
         if ( detector[iarray][idet].y + detector[iarray][idet].r > grid_y_high )
            grid_y_high = detector[iarray][idet].y + detector[iarray][idet].r;
      }

   grid_nx = Nint_f(ceil(grid_x_high/GRID_SIZE) - floor(grid_x_low/GRID_SIZE));
   grid_x_low = GRID_SIZE * floor(grid_x_low/GRID_SIZE);
   grid_x_high = grid_x_low + GRID_SIZE * grid_nx;
   grid_ny = Nint_f(ceil(grid_y_high/GRID_SIZE) - floor(grid_y_low/GRID_SIZE));
   grid_y_low = GRID_SIZE * floor(grid_y_low/GRID_SIZE);
   grid_y_high = grid_y_low + GRID_SIZE * grid_ny;
   
   nn = grid_nx+1;
   if ( grid_ny >= nn )
      nn = grid_ny+1;

#ifdef DEBUG_IACT
printf("Grid: %3.0f<x<%3.0f (%d bins),  %3.0f<y<%3.0f (%d bins)\n",
      grid_x_low,grid_x_high,grid_nx,grid_y_low,grid_y_high,grid_ny);
fflush(stdout);
#endif

   /* Allocate grid and references to detectors on grid elements. */

   grid_elements = 0;
   if ( (grid = (struct gridstruct *) calloc((size_t)(nn*nn)+EXTRA_MEM_10,
           sizeof(struct gridstruct))) == NULL )
   {
      fprintf(stderr,"Grid allocation failed\n");
      return -1;
   }
   grid_elements = nn*nn;

   for (iarray=0; iarray < narray; iarray++)
      for (idet=0; idet<ndet[iarray]; idet++)
      {
         double x0, y0, x1, y1;
         x0 = detector[iarray][idet].x - detector[iarray][idet].r;
         x1 = detector[iarray][idet].x + detector[iarray][idet].r;
         y0 = detector[iarray][idet].y - detector[iarray][idet].r;
         y1 = detector[iarray][idet].y + detector[iarray][idet].r;
         for (ix=(int)((x0-grid_x_low)/GRID_SIZE);
              ix<=(int)((x1-grid_x_low)/GRID_SIZE); ix++)
            for (iy=(int)((y0-grid_y_low)/GRID_SIZE);
                 iy<=(int)((y1-grid_y_low)/GRID_SIZE); iy++)
               if ( ix >= 0 && ix < grid_nx && iy >= 0 && iy < grid_ny )
                  grid[iy*grid_nx+ix].ndet++;
         detector[iarray][idet].iarray = iarray;
         detector[iarray][idet].idet = idet;
      }

   for (iy=0; iy<grid_ny; iy++)
      for (ix=0; ix<grid_nx; ix++)
      {
         struct gridstruct *gc;
         if ( iy*grid_nx+ix < 0 || iy*grid_nx+ix >= grid_elements )
         {
            fprintf(stderr,
           "Grid bounds exceeded*: ix=%d, iy=%d, i=%d (nx=%d, ny=%d, n=%d)\n",
               ix, iy, iy*grid_nx+ix, grid_nx, grid_ny, grid_elements);
            return -1;
         }
         gc = &grid[iy*grid_nx+ix];
         if ( gc->ndet <= 0 )
            gc->detectors = NULL;
         else if ( (gc->detectors = (struct detstruct **)
                 calloc((size_t)gc->ndet+EXTRA_MEM_11,
                    sizeof(struct detstruct *))) == NULL )
	 {
	    fprintf(stderr,"Grid element allocation failed\n");
            return -1;
	 }
      }

   for (iarray=0; iarray < narray; iarray++)
      for (idet=0; idet<ndet[iarray]; idet++)
      {
         double x0, y0, x1, y1;
         x0 = detector[iarray][idet].x - detector[iarray][idet].r;
         x1 = detector[iarray][idet].x + detector[iarray][idet].r;
         y0 = detector[iarray][idet].y - detector[iarray][idet].r;
         y1 = detector[iarray][idet].y + detector[iarray][idet].r;
         for (ix=(int)((x0-grid_x_low)/GRID_SIZE);
              ix<=(int)((x1-grid_x_low)/GRID_SIZE); ix++)
            for (iy=(int)((y0-grid_y_low)/GRID_SIZE);
                 iy<=(int)((y1-grid_y_low)/GRID_SIZE); iy++)
               if ( ix >= 0 && ix < grid_nx && iy >= 0 && iy < grid_ny )
               {
                  struct gridstruct *gc;
                  if ( iy*grid_nx+ix < 0 || iy*grid_nx+ix >= grid_elements )
                  {
                     fprintf(stderr,
                    "Grid bounds exceeded**: ix=%d, iy=%d, i=%d (nx=%d, ny=%d, n=%d)\n",
                        ix, iy, iy*grid_nx+ix, grid_nx, grid_ny, grid_elements);
                     return -1;
                  }
                  gc = &grid[iy*grid_nx+ix];
                  if ( gc->detectors == NULL )
                     fprintf(stderr,"Grid cell problem. No detectors expected.\n");
                  else if ( gc->idet < 0 )
                     fprintf(stderr,"Bad detector count in grid cell.\n");
                  else if ( gc->idet < gc->ndet )
                     gc->detectors[gc->idet++] = &detector[iarray][idet];
                  else
                     fprintf(stderr,"No more detectors added to grid cell.\n");
               }
               else
                  fprintf(stderr,"Outside grid limits: ix=%d, iy=%d, nx=%d, ny=%d\n",
                     ix, iy, grid_nx, grid_ny);
      }

   return 0;
}

#ifdef NO_EXTERN_SAMPLING

/* ------------------------- sample_offset ---------------------------- */
/**
 *  @short Get uniformly sampled or importance sampled offset of array
 *         with respect to core, in the plane perpendicular to the shower axis.
 *
 *  @param sampling_fname Name of file with parameters, to be read on first call.
 *  @param core_range     Maximum core distance as used in data format check [cm].
 *                        If not obeying this maximum distance, make sure to switch on
 *                        the long data format manually.
 *  @param theta          Zenith angle [radians]
 *  @param phi            Shower azimuth angle in CORSIKA angle convention [radians].
 *  @param thetaref       Reference zenith angle (e.g. of VIEWCONE centre) [radians].
 *  @param phiref         Reference azimuth angle (e.g. of VIEWCONE centre) [radians].
 *  @param offax          Angle between central direction (typically VIEWCONE centre)
 *                        and the direction of the current primary [radians].
 *  @param E              Energy of primary particle [GeV]
 *  @param primary        Primary particle ID.
 *  @param xoff           X offset [cm] to be generated.
 *  @param yoff           Y offset [cm] to be generated.
 *  @param sampling_area  Area weight of the generated sample 
 *                        (normalized to Pi*core_range^2) [cm^2].
 */

void sample_offset (char *sampling_fname, double core_range, 
   double theta, double phi, 
   double thetaref, double phiref, double offax, 
   double E, int primary,
   double *xoff, double *yoff, double *sampling_area)
{
   static int init_done = 0;
   double R, p;
   
   if ( !init_done )
   {
      FILE *f = fileopen(sampling_fname,"r");
      if ( f == NULL )
      {
         perror(sampling_fname);
         exit(1);
      }
      /* Read sampling parameters */
      /* ... (TO BE DONE) ... */
      fprintf(stderr,"Sampling parameter file '%s' opened but not used.\n",
         sampling_fname);
      fileclose(f);
      init_done = 1;
   }
   
#ifndef TEST_SAMPLING
   /* In the absence of a real implementation, use uniform distribution. */
   R = core_range*sqrt(rndm(0));
   p = (2.*M_PI) * rndm(1);
   *xoff = R * cos(p);
   *yoff = R * sin(p);
   *sampling_area = M_PI*(core_range*core_range);
#else
   /* Test uniform per radius, not uniform per area. */
   R = core_range*rndm(0);
   p = (2.*M_PI) * rndm(1);
   *xoff = R * cos(p);
   *yoff = R * sin(p);
   *sampling_area = 2.*M_PI*(core_range*R);
#endif
}

#endif /* NO_EXTERN_SAMPLING */
